import os
import importlib
import re
import site
import sys
import uuid
import warnings
from enum import Enum
from pathlib import Path
from typing import Optional, List

try:
    from filelock import FileLock
except ModuleNotFoundError:
    FileLock = None  # type: ignore

import solara.util
from solara.minisettings import BaseSettings

from .. import (  # noqa  # sidefx is that this module creates the ~/.solara directory
    settings,
)
from ..util import get_solara_home

if site.getuserbase() and __file__.startswith(site.getuserbase()):
    prefix = site.getuserbase()
else:
    prefix = sys.prefix


class ThemeVariant(str, Enum):
    light = "light"
    dark = "dark"
    auto = "auto"


class ThemeSettings(BaseSettings):
    variant: ThemeVariant = ThemeVariant.light
    loader: str = "solara"
    show_banner: bool = True
    title: str = "Solara ☀️"

    class Config:
        env_prefix = "solara_theme_"
        case_sensitive = False
        env_file = ".env"


class SSG(BaseSettings):
    # the first app create will initialize this if it is not set
    build_path: Optional[Path] = None
    enabled: bool = False
    headed: bool = False

    class Config:
        env_prefix = "solara_ssg_"
        case_sensitive = False
        env_file = ".env"


class Search(BaseSettings):
    enabled: bool = False

    class Config:
        env_prefix = "solara_search_"


class Telemetry(BaseSettings):
    mixpanel_token: str = "91845eb13a68e3db4e58d64ad23673b7"
    mixpanel_enable: bool = True
    server_user_id: str = "not_set"
    server_fingerprint: str = str(uuid.getnode())
    server_session_id: str = str(uuid.uuid4())

    class Config:
        env_prefix = "solara_telemetry_"
        case_sensitive = False
        env_file = ".env"


class Assets(BaseSettings):
    proxy_cache_dir: Path = Path(prefix + "/share/solara/cdn/")
    fontawesome_enabled: bool = True
    fontawesome_path: str = "/font-awesome@4.5.0/css/font-awesome.min.css"
    extra_locations: List[str] = []

    def extra_paths(self) -> List[Path]:
        # translate locations (packages, directories) into list of paths
        paths = []
        for location in self.extra_locations:
            if Path(location).exists():
                paths.append(Path(location))
            else:
                try:
                    package = importlib.import_module(location)
                except ModuleNotFoundError:
                    raise RuntimeError(f"Could not find {location} as a file or package (SOLARA_ASSETS_EXTRA_LOCATION={self.extra_locations!r}) ")
                if not hasattr(package, "__path__"):
                    raise RuntimeError(f"{location} is not a package (SOLARA_ASSETS_EXTRA_LOCATION={self.extra_locations!r}) ")
                paths.append(Path(package.__path__[0]))
        return paths

    class Config:
        env_prefix = "solara_assets_"
        case_sensitive = False
        env_file = ".env"


class Kernel(BaseSettings):
    cull_timeout: str = "24h"
    max_count: Optional[int] = None
    threaded: bool = solara.util.has_threads

    class Config:
        env_prefix = "solara_kernel_"
        case_sensitive = False
        env_file = ".env"


AUTH0_TEST_CLIENT_ID = "cW7owP5Q52YHMZAnJwT8FPlH2ZKvvL3U"
AUTH0_TEST_CLIENT_SECRET = "zxITXxoz54OjuSmdn-PluQgAwbeYyoB7ALlnLoodftvAn81usDXW0quchvoNvUYD"
AUTH0_TEST_API_BASE_URL = "dev-y02f2bpr8skxu785.us.auth0.com"
AUTH0_LOGOUT_PATH = "v2/logout"

FIEF_TEST_CLIENT_ID = "x2np62qgwp6hnEGTP4JYUE3igdZWhT-AvjpjwwDyKXU"
FIEF_TEST_CLIENT_SECRET = "XQlByE1pVIz5h2SBN2GYDwT_ziqArHJgLD3KqMlCHjg"
FIEF_TEST_API_BASE_URL = "solara-dev.fief.dev"
FIEF_LOGOUT_PATH = "logout"
SESSION_SECRET_KEY_DEFAULT = "change me"

OAUTH_TEST_CLIENT_IDs = [AUTH0_TEST_CLIENT_ID, FIEF_TEST_CLIENT_ID]


class Session(BaseSettings):
    secret_key: str = SESSION_SECRET_KEY_DEFAULT
    http_only: bool = False
    https_only: Optional[bool] = None
    same_site: str = "lax"

    class Config:
        env_prefix = "solara_session_"
        case_sensitive = False
        env_file = ".env"


class OAuth(BaseSettings):
    private: bool = False

    client_id: str = AUTH0_TEST_CLIENT_ID
    client_secret: str = AUTH0_TEST_CLIENT_SECRET
    api_base_url: str = AUTH0_TEST_API_BASE_URL
    logout_path: str = AUTH0_LOGOUT_PATH
    scope: str = "openid profile email"

    class Config:
        env_prefix = "solara_oauth_"
        case_sensitive = False
        env_file = ".env"


HOST_DEFAULT = os.environ.get("HOST", "localhost")
is_mac_os_conda = "arm64-apple-darwin" in HOST_DEFAULT
is_wsl_windows = re.match(r".*?-w1[0-9]", HOST_DEFAULT)
if is_mac_os_conda or is_wsl_windows:
    HOST_DEFAULT = "localhost"


class Server(BaseSettings):
    ignore_nbextensions: List[str] = []

    class Config:
        env_prefix = "solara_server_"
        case_sensitive = False
        env_file = ".env"


class MainSettings(BaseSettings):
    use_pdb: bool = False
    mode: str = "production"
    tracer: bool = False
    timing: bool = False
    root_path: Optional[str] = None  # e.g. /myapp (without trailing slash)
    base_url: str = ""  # e.g. https://myapp.solara.run/myapp/
    platform: str = sys.platform
    host: str = HOST_DEFAULT
    experimental_performance: bool = False

    class Config:
        env_prefix = "solara_"
        case_sensitive = False
        env_file = ".env"


main = MainSettings()
server = Server()
theme = ThemeSettings()
telemetry = Telemetry()
ssg = SSG()
search = Search()
assets = Assets()
oauth = OAuth()
session = Session()
kernel = Kernel()
# fail early
solara.util.parse_timedelta(kernel.cull_timeout)

if settings.assets.proxy:
    try:
        assets.proxy_cache_dir.mkdir(exist_ok=True, parents=True)
    except OSError as e:
        settings.assets.proxy = False
        warnings.warn(
            f"Could not create {assets.proxy_cache_dir} due to {e}. We will automatically disable the assets proxy for you. "
            "If you want to disable this warning, set SOLARA_ASSETS_PROXY to False (e.g. export SOLARA_ASSETS_PROXY=false). "
            "Or change the SOLARA_PROXY_CACHE_DIR environment variable to a directory where you have write access."
        )
        # that's ok, the user probably doesn't have permission to create the directory
        # in this case, we would need to install solara-assets?
        pass

if telemetry.server_user_id == "not_set" and FileLock is not None:
    home = get_solara_home()
    server_user_id_file = home / "server_user_id.txt"
    try:
        with FileLock(str(server_user_id_file) + ".lock"):
            if not server_user_id_file.exists():
                server_user_id_file.write_text(str(uuid.uuid4()))
            telemetry.server_user_id = server_user_id_file.read_text()
    except OSError:
        pass  # it's ok

if oauth.client_id:
    if oauth.client_id not in OAUTH_TEST_CLIENT_IDs:
        if session.secret_key == SESSION_SECRET_KEY_DEFAULT:
            raise ValueError(
                "You must set a session secret key for oauth, it is not safe to use the default. Please set SOLARA_SESSION_SECRET_KEY in your environment."
            )
        if session.https_only is None:
            raise ValueError(
                "You must set https_only for the session to True (recommended) or False in your are using your own oauth provider."
                "Please set SOLARA_SESSION_HTTPS_ONLY in your environment to True or False"
            )
    else:
        # for the test accounts, this is fine
        if session.https_only is None:
            session.https_only = False


# call early so a misconfiguration fails early
assets.extra_paths()



import dataclasses
import importlib.util
import logging
import os
import pickle
import sys
import threading
import traceback
import warnings
import weakref
import html
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional, cast

import ipywidgets as widgets
import reacton
from reacton.core import Element, render
from reacton import ipywidgets

import solara
import solara.lifecycle
from solara.util import nested_get

from . import kernel_context, patch, reload, settings
from .kernel import Kernel
from .utils import pdb_guard

WebSocket = Any
apps: Dict[str, "AppScript"] = {}
thread_lock = threading.RLock()

logger = logging.getLogger("solara.server.app")

reload.reloader.start()


class AppType(str, Enum):
    SCRIPT = "script"
    NOTEBOOK = "notebook"
    MODULE = "module"
    DIRECTORY = "directory"


def display(*args, **kwargs):
    print("display not implemented", args, kwargs)  # noqa


class AppScript:
    directory: Path
    routes: List[solara.Route]

    def __init__(self, name, default_app_name="Page"):
        self.fullname = name
        if reload.reloader.on_change:
            raise RuntimeError("Previous reloader still had a on_change attached, no cleanup?")
        reload.reloader.on_change = self.on_file_change

        self.app_name = default_app_name
        if ":" in self.fullname:
            self.name, self.app_name = self.fullname.rsplit(":", 1)
            if len(self.name) == 1:  # a windows drive letter, restore
                self.name = self.fullname
                self.app_name = default_app_name
        else:
            self.name = name
        self.path: Path = Path(self.name).resolve()
        if self.path.is_dir():
            self.type = AppType.DIRECTORY
            # resolve the directory, because Path("file").parent.parent == "." != ".."
            self.directory = self.path.resolve()
        elif self.name.endswith(".py"):
            self.type = AppType.SCRIPT
            self.directory = self.path.parent.resolve()
        elif self.name.endswith(".ipynb"):
            self.type = AppType.NOTEBOOK
            self.directory = self.path.parent.resolve()
        else:
            self.type = AppType.MODULE
            try:
                spec = importlib.util.find_spec(self.name)
            except ValueError:
                if self.name not in sys.modules:
                    raise ImportError(f"Module {self.name} not found")
                spec = importlib.util.spec_from_file_location(self.name, sys.modules[self.name].__file__)
            if spec is None:
                raise ImportError(f"Module {self.name} cannot be found")
            assert spec is not None
            if spec.origin is None:
                raise ImportError(f"Module {self.name} cannot be found, or is a namespace package")
            assert spec.origin is not None
            self.path = Path(spec.origin)
            self.directory = self.path.parent
        self._initialized = False
        self._lock = threading.Lock()

    def init(self):
        try:
            context = kernel_context.get_current_context()
        except RuntimeError:
            context = None
        if context is not None:
            raise RuntimeError(f"We should not have an existing Solara app context when running an app for the first time: {context}")
        dummy_kernel_context = kernel_context.create_dummy_context()
        with dummy_kernel_context:
            app = self._execute()

        # We now ran the app, now we can check for patches that require heavy imports
        patch.patch_heavy_imports()

        self._first_execute_app = app
        reload.reloader.root_path = self.directory
        if self.type == AppType.MODULE:
            package_name = self.name.split(".")[0]
            mod = importlib.import_module(package_name)
            if mod.__file__ is not None:
                package_root_path = Path(mod.__file__).parent
                reload.reloader.root_path = package_root_path
        dummy_kernel_context.close()
        self._initialized = True

    def _execute(self):
        logger.info("Executing %s", self.name)
        app = None
        routes: Optional[List[solara.Route]] = None

        def add_path():
            # this is not expected for modules, similar to `python script.py and python -m package.mymodule`
            if self.type in [AppType.SCRIPT, AppType.NOTEBOOK]:
                working_directory = str(self.path.parent)
                if working_directory not in sys.path:
                    sys.path.insert(0, working_directory)

        if self.type == AppType.DIRECTORY:
            # resolve the directory, because Path("file").parent.parent == "." != ".."
            routes = solara.generate_routes_directory(self.path)

            if any(name for name in sys.modules.keys() if name.startswith(self.name)):
                logger.warning(
                    f"Directory {self.name} is also used as a package. This can cause modules to be loaded twice, and might "
                    "cause unexpected behavior. If you run solara from a different directory (e.g. the parent directory) you "
                    "can avoid this ambiguity."
                )

        elif self.type == AppType.SCRIPT:
            add_path()
            # manually add the script to the watcher
            reload.reloader.watcher.add_file(self.path)
            initial_namespace = {
                "__name__": "__main__",
            }
            with reload.reloader.watch():
                routes = [solara.autorouting._generate_route_path(self.path, first=True, initial_namespace=initial_namespace)]
        elif self.type == AppType.NOTEBOOK:
            add_path()
            # manually add the notebook to the watcher
            reload.reloader.watcher.add_file(self.path)
            with reload.reloader.watch():
                routes = [solara.autorouting._generate_route_path(self.path, first=True)]
        else:
            # the module itself will be added by reloader
            # automatically
            with kernel_context.without_context(), reload.reloader.watch():
                self.type = AppType.MODULE
                mod = importlib.import_module(self.name)
                routes = solara.generate_routes(mod)

        app = solara.autorouting.RenderPage(self.app_name)

        # when the root moduled defined routes, skip the enclosing route object
        if len(routes) == 1 and routes[0].module and hasattr(routes[0].module, "routes"):
            if routes[0].component:
                warnings.warn(
                    f"{self.name} has a component defined, but you are also defining routes."
                    " To avoid confusing, consider renaming the {self.app_name} component."
                )
            routes = routes[0].children

        if self.app_name != "Page":
            # if we specified the app name, we replace the component
            if len(routes) > 1:
                raise ValueError(f"App {self.name} has multiple routes, but a default app name was given: {self.app_name}")
            assert len(routes) == 1
            component = nested_get(routes[0].module.__dict__, self.app_name, None)
            routes = [dataclasses.replace(routes[0], component=component)]

        if settings.ssg.build_path is None:
            settings.ssg.build_path = self.directory.parent.resolve() / "build"

        # auto enable search if search.json exists
        search_index_file = self.directory.parent / "assets" / "search.json"
        if search_index_file.exists():
            settings.search.enabled = True

        # this might be useful for development
        # but requires reloading of react in solara itself
        # for name, module in sys.modules.items():
        #     if name.startswith("reacton"):
        #         file = inspect.getfile(module)
        #         self.watcher.add_file(file)

        # cgi vars: https://datatracker.ietf.org/doc/html/rfc3875
        # we cannot set script name, because gunicorn uses it (and will crash)
        # os.environ["SCRIPT_NAME"] = self.name
        os.environ["PATH_TRANSLATED"] = str(self.path.resolve())

        self.routes = routes

        # this might be useful for development
        # but requires reloading of react in solara itself
        # for name, module in sys.modules.items():
        #     if name.startswith("reacton"):
        #         file = inspect.getfile(module)
        #         self.watcher.add_file(file)

        # cgi vars: https://datatracker.ietf.org/doc/html/rfc3875
        # we cannot set script name, because gunicorn uses it (and will crash)
        # os.environ["SCRIPT_NAME"] = self.name
        os.environ["PATH_TRANSLATED"] = str(self.path.resolve())
        return app

    def close(self):
        reload.reloader.on_change = None
        context_values = list(kernel_context.contexts.values())
        kernel_context.contexts.clear()
        for context in context_values:
            context.close()

    def check(self):
        if not self._initialized:
            with self._lock:
                if not self._initialized:
                    self.init()

    def run(self):
        self.check()
        if self._first_execute_app is None:
            with thread_lock:
                if self._first_execute_app is None:
                    try:
                        self._first_execute_app = self._execute()
                        print("Re-executed app", self.name)  # noqa
                    except Exception as e:
                        error = ""
                        error = "".join(traceback.format_exception(None, e, e.__traceback__))
                        print(error, file=sys.stdout, flush=True)  # noqa

                        error = html.escape(error)
                        self._first_execute_app = ipywidgets.HTML(value=f"<pre>{error}</pre>", layout=ipywidgets.Layout(overflow="auto"))
                        # We now ran the app again, might contain new imports

                        print("Failed to execute app, fix the error and save the file to reload")  # noqa
                    # We now ran the app again, might contain new imports
                    patch.patch_heavy_imports()

        return self._first_execute_app

    def on_file_change(self, name):
        path = Path(name)
        if path.suffix == ".vue":
            logger.info("Vue file changed: %s", name)
            template_content = path.read_text(encoding="utf-8")
            for context in list(kernel_context.contexts.values()):
                with context:
                    for filepath, widget in context.templates.items():
                        if filepath == str(path):
                            widget.template = template_content
        else:
            logger.info("Reload requires due to change in module: %s", name)
            self.reload()

    def reload(self):
        # if multiple files change in a short time, we want to do this
        # not concurrently. Even better would be to do a debounce?
        with thread_lock:
            # TODO: clearing the type_counter is a bit of a hack
            # and we should introduce reload 'hooks', so there is
            # less interdependency between modules
            import solara.lab.toestand

            solara.lab.toestand.ConnectionStore._type_counter.clear()

            # we need to remove callbacks that are added in the app code
            # which will be re-executed after the reload and we do not
            # want to keep executing the old ones.
            for kc in solara.lifecycle._on_kernel_start_callbacks.copy():
                callback, path, module, cleanup = kc
                will_reload = False
                if module is not None:
                    module_name = module.__name__
                    if module_name in reload.reloader.get_reload_module_names():
                        will_reload = True
                elif path is not None:
                    if str(path.resolve()).startswith(str(self.directory)):
                        will_reload = True
                    else:
                        logger.warning(
                            "script %s is not in the same directory as the app %s but is using on_kernel_start, "
                            "this might lead to multiple entries, and might indicate a bug.",
                            path,
                            self.directory,
                        )

                if will_reload:
                    logger.info("reload: Removing on_kernel_start callback: %s (since it will be added when reloaded)", callback)
                    cleanup()

            context_values = list(kernel_context.contexts.values())
            # save states into the context so the hot reload will
            # keep the same state
            for context in context_values:
                render_context = cast(reacton.core._RenderContext, context.app_object)
                if render_context:
                    with context:
                        # we save the state for when the app reruns, so we stay in the same state.
                        # (e.g. button clicks, chosen options etc)
                        # for instance a dataframe, needs to be pickled, because after the pandas
                        # module is reloaded, it's a different class type
                        logger.info("pickling state: %s", render_context.state_get())
                        try:
                            context.state = pickle.dumps(render_context.state_get())
                        except Exception as e:
                            logger.warning("Could not pickle state, next render the state will be lost: %s", e)
                        # clear/cleanup the render_context, so during reload we start
                        # from scratch
                        context.app_object = None
                        # we want to reuse the container
                        render_context.container = None
                        try:
                            render_context.close()
                        except Exception as e:
                            logger.exception("Could not close render context: %s", e)

            self._first_execute_app = None
            self.run()

            # ask all contexts/users to reload
            for context in context_values:
                with context:
                    context.reload()


def _run_app(
    app_state,
    app_script: AppScript,
    pathname: str,
    render_context: reacton.core._RenderContext = None,
):
    # app.signal_hook_install()
    main_object = app_script.run()
    app_state = pickle.loads(app_state) if app_state is not None else None
    if app_state:
        logger.info("Restoring state: %r", app_state)

    context = kernel_context.get_current_context()
    container = context.container
    if isinstance(main_object, widgets.Widget):
        return main_object, render_context
    elif isinstance(main_object, Element) or isinstance(main_object, reacton.core.Component):
        if isinstance(main_object, Element):
            children = [main_object]
        else:
            children = [main_object()]
        solara_context = solara.RoutingProvider(children=children, routes=app_script.routes, pathname=pathname)
        if render_context is None:
            result = render(solara_context, container, handle_error=True, initial_state=app_state)
            # support older versions of react
            if isinstance(result, tuple):
                container, render_context = result
            else:
                render_context = result
        else:
            if app_state:
                render_context.state_set(render_context.context_root, app_state)
            result = render_context.render(solara_context)
            container = render_context.container
        # return container, render_context
    else:
        extra = ""
        dotted = []
        for key, value in vars(main_object).items():
            if isinstance(value, (Element, widgets.Widget)):
                dotted.append(f"{app_script.app_name}.{key}")
        if dotted:
            extra = " We did find that sub objects that might work: " + ", ".join(dotted)
        raise ValueError(
            f"Main object (with name {app_script.app_name} in {app_script.path}) is not a Widget, Element or Component, but {type(main_object)}." + extra
        )
    return container, render_context


def load_app_widget(app_state, app_script: AppScript, pathname: str):
    # load the app, and set it at the child of the context's container
    app_state_initial = app_state
    context = kernel_context.get_current_context()
    container = context.container
    assert container is not None
    try:
        import ipyreact

        del ipyreact
    except ModuleNotFoundError:
        pass
    else:
        import solara.server.esm

        # will create widgets, but will clean itself up when the kernel closes
        solara.server.esm.create_modules()
        solara.server.esm.create_import_map()

    try:
        render_context = context.app_object
        app_state = app_state_initial
        with pdb_guard():
            widget, render_context = _run_app(
                app_state,
                app_script,
                pathname,
                render_context=render_context,
            )
            if render_context is None:
                assert context.container is not None
                context.container.children = [widget]

        if render_context:
            context.app_object = render_context

    except BaseException as e:
        error = ""
        error = "".join(traceback.format_exception(None, e, e.__traceback__))
        print(error, file=sys.stdout, flush=True)  # noqa
        # widget = widgets.Label(value="Error, see server logs")
        import html

        error = html.escape(error)
        with context:
            widget = widgets.HTML(f"<pre>{error}</pre>", layout=widgets.Layout(overflow="auto"))
            container.children = [widget]


def load_themes(themes: Dict[str, Dict[str, str]], dark: bool):
    # While these usually gets set from the frontend, in solara (server) we want to know theme information directly at the first
    # render. Also, using the same trait allows us to write code which works on all widgets platforms, instead
    # or using something different when running under solara server
    from solara.lab.components.theming import _set_theme, theme

    _set_theme(themes)
    theme.dark_effective = dark


def solara_comm_target(comm, msg_first):
    app: Optional[AppScript] = None

    def on_msg(msg):
        nonlocal app
        comm = comm_ref()
        assert comm is not None
        context = kernel_context.get_current_context()
        data = msg["content"]["data"]
        method = data["method"]
        if method == "run":
            args = data["args"]
            path = args.get("path", "")
            app_name = args.get("appName") or "__default__"
            app = apps[app_name]
            context = kernel_context.get_current_context()
            import ipyvuetify

            container = ipyvuetify.Html(tag="div")
            context.container = container
            themes = args.get("themes")
            dark = args.get("dark")
            load_themes(themes, dark)
            try:
                load_app_widget(None, app, path)
            except Exception as e:
                msg = f"Error loading app: from path {path} and app {app_name}"
                logger.exception(msg)
                raise RuntimeError(msg) from e
            comm.send({"method": "finished", "widget_id": context.container._model_id})
        elif method == "app-status":
            context = kernel_context.get_current_context()
            # if there is no container, we never ran the app
            if context.container is not None:
                logger.info("app-status check: %s app started", context.id)
                comm.send({"method": "app-status", "started": True})
            else:
                logger.info("app-status check: %s app not started", context.id)
                comm.send({"method": "app-status", "started": False})

        elif method == "reload":
            from solara.lab.components.theming import _get_theme, theme

            assert app is not None
            context = kernel_context.get_current_context()
            path = data.get("path", "")
            current_theme = theme._instance.value
            theme_dict = _get_theme(current_theme)

            with context:
                context.restart()
                load_themes(theme_dict, current_theme.dark_effective)
                load_app_widget(context.state, app, path)
                comm.send({"method": "finished"})
        else:
            logger.error("Unknown comm method called on solara.control comm: %s", method)

    def reload():
        comm = comm_ref()
        assert comm is not None
        context = kernel_context.get_current_context()
        # we don't reload the app ourself, we send a message to the client
        # this ensures that we don't run code of any client that for some reason is connected
        # but not working anymore. And it indirectly passes a message from the current thread
        # (which is that of the Reloader/watchdog), to the thread of the client
        logger.debug(f"Send reload to client: {context.id}")
        comm.send({"method": "reload"})

    comm.on_msg(on_msg)
    comm_ref = weakref.ref(comm)
    del comm

    kernel_context.get_current_context().reload = reload


def register_solara_comm_target(kernel: Kernel):
    kernel.comm_manager.register_target("solara.control", solara_comm_target)


from . import patch  # noqa

patch.patch()
# the default app (used in solara-server)
if "SOLARA_APP" in os.environ:
    with pdb_guard():
        apps["__default__"] = AppScript(os.environ.get("SOLARA_APP", "solara.website.pages:Page"))


@solara.util.once
def ensure_apps_initialized():
    for app in apps.values():
        app.init()



import time
from pathlib import Path
from typing import Optional, cast

import solara
from solara.alias import rw

HERE = Path(__file__).parent
title = "use_thread"
__doc__ = open(HERE / "use_thread.md").read()


@solara.component
def Page():
    number, set_number = solara.use_state(17)
    # the number that proofs it is not a prime
    proof, set_proof = solara.use_state(cast(Optional[int], None))

    def work():
        for i in range(3, number):
            reminder = number % i
            if reminder == 0:
                set_proof(i)
                return False
            # make it always take ~4 seconds
            time.sleep(4 / number)
        return True

    # work will be cancelled/restarted every time the dependency changes
    result: solara.Result[bool] = solara.use_thread(work, dependencies=[number])

    rw.IntText(value=number, on_value=set_number)
    if result.state == solara.ResultState.FINISHED:
        if result.value:
            solara.Success(f"{number} is a prime!")
        else:
            solara.Error(f"{number} is not a prime, it can be divided by {proof} ")
    elif result.state == solara.ResultState.ERROR:
        solara.Error(f"Error occurred: {result.error}")
    else:
        solara.Info(f"Running... (status = {result.state})")
        solara.v.ProgressLinear(indeterminate=True)



from jupyter_server.utils import url_path_join

from solara.server.cdn_helper import cdn_url_path
from solara.server.jupyter.cdn_handler import CdnHandler
from .solara import SolaraHandler, Assets, ReadyZ


def _jupyter_server_extension_paths():
    return [{"module": "solara.server.jupyter.server_extension"}]


def _load_jupyter_server_extension(server_app):
    # a dummy app, so that server.read_root can be used
    import solara.server.app

    solara.server.app.apps["__default__"] = solara.server.app.AppScript("solara.server.jupyter.solara:Page")
    solara.server.app.apps["__default__"].init()

    web_app = server_app.web_app

    host_pattern = ".*$"
    base_url = url_path_join(web_app.settings["base_url"])

    web_app.add_handlers(
        host_pattern,
        [
            (url_path_join(base_url, f"/{cdn_url_path}/(.*)"), CdnHandler, {}),  # kept for backward compatibility
            (url_path_join(base_url, f"/solara/{cdn_url_path}/(.*)"), CdnHandler, {}),
            (url_path_join(base_url, "/solara/static/assets/(.*)"), Assets, {}),
            (url_path_join(base_url, "/solara/readyz"), ReadyZ, {}),
            (url_path_join(base_url, "/solara(.*)"), SolaraHandler, {}),
        ],
    )


# For backward compatibility
load_jupyter_server_extension = _load_jupyter_server_extension

# For future compatibility
_jupyter_server_extension_points = _jupyter_server_extension_paths



"""
# use_reactive

"""

import solara
from solara.website.components import NoPage
from solara.website.utils import apidoc

title = "use_reactive"


Page = NoPage


__doc__ += apidoc(solara.use_reactive)  # type: ignore



import json
import logging
import os
import platform
import threading
import time
import uuid
from collections import defaultdict
from typing import Dict, Optional
from urllib.parse import quote

import ipyvue
import ipyvuetify
import ipywidgets
import requests

import solara
import solara.util

from . import settings

logger = logging.getLogger("solara.server.telemetry")

_auto_restart_enabled = False
_server_user_id_override = None
_server_start_time = time.time()
# Privacy note: mixpanel does not store the IP, only the region
_server_ip = None
_platform_system = platform.system()
_platform_release = platform.release()
_python_version = platform.python_version()
_connections_per_session_daily: Dict[str, int] = defaultdict(int)
_connections_per_session_cumulative: Dict[str, int] = defaultdict(int)

_seconds_per_day = 60 * 60 * 24
_report_timeout = _seconds_per_day

solara_props = {
    "solara_version": solara.__version__,
    "ipywidgets_version": ipywidgets.__version__,
    "ipyvuetify_version": ipyvuetify.__version__,
    "ipyvue_version": ipyvue.__version__,
}
_docker = False
_compute_platform = "unknown"


def is_running_on_domino():
    return "DOMINO_PROJECT_OWNER" in os.environ


def is_running_on_sagemaker_notebook():
    return "AWS_EXECUTION_ENV" in os.environ and "SageMaker" in os.environ["AWS_EXECUTION_ENV"]


def is_running_on_azure():
    return "AZURE_NOTEBOOKS_VM" in os.environ or "AZUREML_RUN_ID" in os.environ


if solara.util.is_running_in_colab():
    _compute_platform = "colab"
elif is_running_on_domino():
    _compute_platform = "domino"
elif is_running_on_sagemaker_notebook():
    _compute_platform = "sagemaker"
elif is_running_on_azure():
    _compute_platform = "azure"

_vscode = solara.util.is_running_in_vscode()

try:
    path = "/proc/self/cgroup"
    _docker = os.path.exists("/.dockerenv") or os.path.isfile(path) and any("docker" in line for line in open(path))
except:  # noqa
    logger.exception("Failed to detect docker")


def _get_ip():
    global _server_ip
    try:
        _server_ip = requests.get("https://api.ipify.org").text
    except Exception:
        _server_ip = "failed to get IP"


def override_server_user_id(server_user_id: str):
    global _server_user_id_override
    _server_user_id_override = server_user_id


def get_server_user_id():
    return _server_user_id_override or settings.telemetry.server_user_id


def track(event: str, props: Optional[Dict] = None):
    if _auto_restart_enabled:
        return
    if not settings.telemetry.mixpanel_enable:
        return
    event_item = {
        "event": event,
        "properties": {
            "token": settings.telemetry.mixpanel_token,
            "fingerprint": settings.telemetry.server_fingerprint,
            "time": int(time.time() * 1000),
            "distinct_id": get_server_user_id(),
            # can be useful to get of session duration
            "session_id": settings.telemetry.server_session_id,
            "$insert_id": str(uuid.uuid4()),  # to de-duplicate events
            # Privacy note: mixpanel does not store the IP, only the region
            "ip": _server_ip,
            "platform_system": _platform_system,
            "platform_release": _platform_release,
            "python_version": _python_version,
            "docker": _docker,
            "compute_platform": _compute_platform,
            "vscode": _vscode,
            "solara_mode": settings.main.mode,
            **(solara_props or {}),
            **(props or {}),
        },
    }
    try:
        requests.post(
            "https://api.mixpanel.com/track/",
            headers={"content-type": "application/x-www-form-urlencoded"},
            data=f"data={quote(json.dumps([event_item]))}",
            timeout=1,
        )
    except Exception:
        pass


def _usage_stats():
    unique_session_daily = len(_connections_per_session_daily)
    unique_session_cumulative = len(_connections_per_session_cumulative)
    avg_connections_daily = sum(_connections_per_session_daily.values()) / unique_session_daily if unique_session_daily else 0
    avg_connections_cumulative = sum(_connections_per_session_cumulative.values()) / unique_session_cumulative if unique_session_cumulative else 0
    uptime_days = (time.time() - _server_start_time) / (_seconds_per_day)
    return {
        "uptime_days": uptime_days,
        "unique_session_daily": unique_session_daily,
        "avg_connections_daily": avg_connections_daily,
        "unique_session_cumulative": unique_session_cumulative,
        "avg_connections_cumulative": avg_connections_cumulative,
    }


def _track():
    global _server_start_time
    _server_start_time = time.time()
    _get_ip()
    track("Solara server start")
    while True:
        try:
            time.sleep(_report_timeout)
            track("Solara report", _usage_stats())
            _connections_per_session_daily.clear()
        except Exception:
            pass


_thread = threading.Thread(target=_track, daemon=True)


def server_start():
    if _auto_restart_enabled:
        return
    if not settings.telemetry.mixpanel_enable:
        return
    try:
        _thread.start()
    except RuntimeError:
        pass


def server_stop():
    duration = time.time() - _server_start_time
    track("Solara server stop", {"duration_seconds": duration, **_usage_stats()})


_jupyter_start_event_send = False


def jupyter_start():
    # track the usage once
    global _jupyter_start_event_send
    if _jupyter_start_event_send:
        return
    _jupyter_start_event_send = True

    def jupyter_track():
        _get_ip()
        track("Solara Jupyter use")

    try:
        threading.Thread(target=jupyter_track, daemon=True).start()
    except:  # noqa
        pass


def connection_open(session_id):
    _connections_per_session_daily[session_id] += 1
    _connections_per_session_cumulative[session_id] += 1


def connection_close(session_id):
    pass


if __name__ == "__main__":
    track("Solara test event", {"where": "command line"})



import os
import sys

import playwright.sync_api
import pytest

pytest.importorskip("solara_enterprise")

from solara_enterprise.auth import get_logout_url  # noqa

from solara.server import settings  # noqa

if sys.version_info[:2] <= (3, 6):
    pytest.skip("Test requires python 3.7 or higher", allow_module_level=True)


@pytest.mark.skipif(not bool(os.environ.get("AUTH0_PASSWORD")), reason="AUTH0_PASSWORD not set")
def test_oauth_from_app_auth0(page_session: playwright.sync_api.Page, solara_server, solara_app):
    with solara_app("solara.website.pages"):
        settings.main.base_url = ""
        settings.oauth.client_id = settings.AUTH0_TEST_CLIENT_ID
        settings.oauth.client_secret = settings.AUTH0_TEST_CLIENT_SECRET
        settings.oauth.api_base_url = settings.AUTH0_TEST_API_BASE_URL
        settings.oauth.logout_path = settings.AUTH0_LOGOUT_PATH
        page_session.goto(solara_server.base_url + "/documentation/examples/general/login_oauth")
        page_session.locator("_vue=v-btn >> text=Login").click()
        page_session.locator('css=input[name="username"]').fill(os.environ["AUTH0_USERNAME"])
        page_session.locator('css=input[name="password"]').fill(os.environ["AUTH0_PASSWORD"])
        page_session.locator('css=button[name="action"]').nth(-1).click()
        page_session.locator("_vue=v-btn >> text=Logout").click()
        # do another round, we've
        page_session.locator("_vue=v-btn >> text=Login").click()
        page_session.locator('css=input[name="username"]').fill(os.environ["AUTH0_USERNAME"])
        page_session.locator('css=input[name="password"]').fill(os.environ["AUTH0_PASSWORD"])
        page_session.locator('css=button[name="action"]').nth(-1).click()
        page_session.locator("_vue=v-btn >> text=Logout").click()
        page_session.locator("_vue=v-btn >> text=Login").wait_for()
        page_session.locator("_vue=v-btn >> text=Login").wait_for()


@pytest.mark.skip(reason="Fief support is deprecated for now")
def test_oauth_from_app_fief(page_session: playwright.sync_api.Page, solara_server, solara_app):
    with solara_app("solara.website.pages"):
        settings.main.base_url = ""
        settings.oauth.client_id = settings.FIEF_TEST_CLIENT_ID
        settings.oauth.client_secret = settings.FIEF_TEST_CLIENT_SECRET
        settings.oauth.api_base_url = settings.FIEF_TEST_API_BASE_URL
        settings.oauth.logout_path = settings.FIEF_LOGOUT_PATH
        page_session.goto(solara_server.base_url + "/documentation/examples/general/login_oauth")
        page_session.locator("_vue=v-btn >> text=Login").click()
        page_session.locator('css=input[name="email"]').fill(os.environ["FIEF_USERNAME"])
        page_session.locator('css=input[name="password"]').fill(os.environ["FIEF_PASSWORD"])
        page_session.locator('css=button[type="submit"]').click()
        page_session.locator("_vue=v-btn >> text=Logout").click()
        page_session.locator("_vue=v-btn >> text=Login").wait_for()


@pytest.mark.skipif(not bool(os.environ.get("AUTH0_PASSWORD")), reason="AUTH0_PASSWORD not set")
def test_oauth_private(page_session: playwright.sync_api.Page, solara_server, solara_app):
    settings.oauth.private = True
    try:
        with solara_app("solara.website.pages"):
            settings.main.base_url = ""
            settings.oauth.client_id = settings.AUTH0_TEST_CLIENT_ID
            settings.oauth.client_secret = settings.AUTH0_TEST_CLIENT_SECRET
            settings.oauth.api_base_url = settings.AUTH0_TEST_API_BASE_URL
            settings.oauth.logout_path = settings.AUTH0_LOGOUT_PATH

            response = page_session.goto(solara_server.base_url + "/static/public/beach.jpeg?cache=off")
            assert response is not None
            assert response.status == 401
            response_invalid = page_session.goto(solara_server.base_url + "/invalid_url")
            assert response_invalid is not None
            assert response_invalid.status == 401
            page_session.goto(solara_server.base_url + "/documentation/components/advanced/style")
            page_session.locator('css=input[name="username"]').fill(os.environ["AUTH0_USERNAME"])
            page_session.locator('css=input[name="password"]').fill(os.environ["AUTH0_PASSWORD"])
            page_session.locator('css=button[name="action"]').nth(-1).click()
            page_session.locator("text=Add a custom piece of CSS").wait_for()
            response = page_session.goto(solara_server.base_url + "/static/public/beach.jpeg")
            assert response is not None
            assert str(response.status)[0] == "2"  # check 2xx
            page_session.goto(get_logout_url("/documentation/components/advanced/style"))
            page_session.locator('css=input[name="username"]').fill(os.environ["AUTH0_USERNAME"])
            response = page_session.goto(solara_server.base_url + "/static/public/beach.jpeg?cache=off")
            assert response is not None
            assert response.status == 401
    finally:
        settings.oauth.private = False



"""# Head"""

import solara
from solara.website.utils import apidoc


@solara.component
def Page():
    solara.Info("A Head component does not render somesome visual on the page, but it is used to avoid duplicate tags, such as titles.")
    with solara.Head():
        # title should always occur inside a Head component
        solara.Title("Custom title")


__doc__ += apidoc(solara.Head.f)  # type: ignore



import os
import playwright.sync_api

TEST_PORT = int(os.environ.get("PORT", "18765"))


def test_pyinstaller_basics(page: playwright.sync_api.Page):
    page.goto(f"http://localhost:{TEST_PORT}")
    page.locator("text=Clicked: 0").click()
    page.locator("text=Clicked: 1").wait_for()



"""Build webapps using IPywidgets"""

__version__ = "1.51.1"
github_url = "https://github.com/widgetti/solara"
git_branch = "master"


from . import comm  # noqa: F401


def _using_solara_server():
    import sys

    if "solara.server.starlette" in sys.modules or "solara.server.flask" in sys.modules:
        return True
    if sys.argv[0].split("/")[-1] == "solara":
        return True
    return False


# isort: skip_file
# ruff: noqa: F401 F403
from reacton import (
    component_interactive,
    value_component,
    create_context,
    get_widget,
    get_context,
    make,
    provide_context,
    render,
    render_fixed,
    use_context,
    use_effect,
    use_exception,
    use_memo,
    use_reducer,
    use_ref,
    use_side_effect,
    use_state,
    use_state_widget,
)
from reacton.core import Element
from .core import component

try:
    import ipyvuetify.components as v  # type: ignore
except ModuleNotFoundError:
    # backwards compatibility
    import reacton.ipyvuetify as v  # type: ignore
from . import util

from .reactive import *

from .datatypes import *
from .hooks import *
from .cache import memoize
from . import cache

# TODO: components re-exports v, we should use __all__ in components/misc.py
from .components import *  # type: ignore
from .components import _component_vue

from .routing import use_route, use_router, use_route_level, find_route, use_pathname, resolve_path
from .autorouting import generate_routes, generate_routes_directory, RenderPage, RoutingProvider, DefaultLayout
from .checks import check_jupyter
from .scope import get_kernel_id, get_session_id


def display(*objs, **kwargs):
    """Display a Python object in the current component.

    ## Supported objects

    Any object that can be displayed in the Jupyter notebook can be displayed in Solara
    as well. This means that plotly, pandas, vaex and other libraries that have
    a display function will work out of the box in Solara.

    However, if you require callback functions, use the specific Solara components, e.g.:

      * [Plotly](/documentation/components/viz/plotly)
      * [Altair](/documentation/components/viz/altair)
      * [Matplotlib](/documentation/components/viz/matplotlib)
      * [Dataframe](/documentation/components/data/dataframe)

    ```solara
    import solara
    import pandas as pd
    import plotly.express as px

    df = px.data.gapminder()
    fig = px.scatter(df, x="gdpPercap", y="lifeExp")
    button = solara.Button("Click me")

    @solara.component
    def Page():

        # You can use solara.display directly
        solara.display(fig)

        # Or use the IPython display function, which is in the global scope
        # in Jupyter and Solara
        display(df)

        with solara.Card("Some card"):
            # you can use display to add existing elements to the parent's
            # children
            display(button)
    ```
    Note that this is a dispatch call to the
    [IPython display function](https://ipython.readthedocs.io/en/stable/api/generated/IPython.display.html#IPython.display.display),
    with a slimmer API. The main purpose of this function is the have type safety, since tools like mypy do now know about
    the existence of the IPython display function, and will complain if you use it without importing it.
    Since solara is always imported, this saves you from having to import IPython in your code.

    ## Arguments

     * `objs`: The objects to display.
     * `kwargs`: Keyword arguments to pass to the IPython display function.

    """
    from IPython.display import display as ipy_display

    ipy_display(*objs, **kwargs)



from typing import Callable

import solara
import solara.lab
from solara.alias import rv
from solara.server import settings
from .algolia import Algolia


@solara.component
def Header(
    on_toggle_left_menu: Callable[[], None] = None,
    on_toggle_right_menu: Callable[[], None] = None,
):
    # use routes of parent (assuming we are a child of a layout)
    route_current, all_routes = solara.use_route(level=-1)
    route_current_with_children, all_routes_with_children = solara.use_route()
    router = solara.use_router()
    dark_effective = solara.lab.use_dark_effective()

    # set states for menu
    with solara.Column(gap="0px"):
        # with solara.Div(classes=["news"]):
        #     solara.HTML(
        #         "div",
        #         unsafe_innerHTML="<a href='https://github.com/widgetti/solara' target='_blank' >Star us on github 🤩</a>",
        #     )
        with solara.v.AppBar(
            tag="header", flat=True, clipped_left=True, style_="background-color: transparent; border-bottom: 1px solid var(--color-border-appbar);"
        ):
            if route_current is not None and route_current.module is not None and hasattr(route_current.module, "Sidebar"):
                with solara.Button(icon=True, class_="hidden-md-and-up", on_click=lambda: on_toggle_left_menu and on_toggle_left_menu()):
                    rv.Icon(children=["mdi-menu"])

            display = " d-none d-sm-flex" if route_current is not None and route_current.path not in ["about", "pricing", "careers"] else " d-flex"
            with solara.v.Html(
                tag="div",
                class_="header-logo-container" + display,
                style_="""
                    background-color: transparent;
                    flex-grow: 1;
                    align-items: stretch;
                    max-height: 65%;
                """,
            ):
                with solara.Link(path_or_route="/", style={"display": "flex", "align-items": "center", "flex-direction": "row", "gap": "10px"}):
                    solara.Image(router.root_path + f"/static/assets/images/logo{'_white' if dark_effective else ''}.svg", classes=["header-logo"])
                    solara.Text("API", style={"font-size": "20px", "font-weight": "600"})

            if route_current is not None and route_current.path not in ["about", "pricing", "careers"]:
                if settings.search.enabled:
                    from solara_enterprise.search.search import Search

                    Search()
                else:
                    Algolia()

            with rv.Html(tag="ul", class_="main-menu menu d-none d-md-flex", style_="justify-content: flex-end;"):
                for route in all_routes:
                    if route.path in ["apps", "contact", "changelog", "our_team", "about", "pricing", "roadmap", "careers"]:
                        continue
                    current = route_current == route
                    with rv.Html(tag="li", class_="active" if current else None):
                        solara.Link("/" + route.path if route.path != "/" else "/", children=[route.label])
            with rv.Btn(icon=True, tag="a", class_="d-none d-md-flex", attributes={"href": solara.github_url, "target": "_blank"}):
                rv.Icon(children=["mdi-github-circle"])

            with rv.Btn(icon=True, tag="a", class_="d-none d-md-flex", attributes={"href": "https://discord.solara.dev", "target": "_blank"}):
                rv.Icon(children=["mdi-discord"])

            with solara.v.Html(tag="div", class_="d-none d-md-flex"):
                solara.lab.ThemeToggle()

            # with solara.Button(icon=True, class_="hidden-md-and-up", on_click=lambda: on_toggle_right_menu and on_toggle_right_menu()):
            #     rv.Icon(children=["mdi-menu"])



from pathlib import Path

import solara
from solara.website.components.mailchimp import MailChimp
from solara.website.components.notebook import Notebook

HERE = Path(__file__).parent
title = "Jupyter Dashboard (1/3)"


@solara.component
def Page():
    with solara.Column(gap=0):
        title = "Build your Jupyter dashboard using Solara"
        solara.Meta(property="og:title", content=title)
        solara.Meta(name="twitter:title", content=title)
        solara.Title(title)

        img = "https://dxhl76zpt6fap.cloudfront.net/public/docs/tutorial/jupyter-dashboard1.webp"
        solara.Meta(name="twitter:image", content=img)
        solara.Meta(property="og:image", content=img)

        description = "Learn how to build a Jupyter dashboard and deploy it as a web app using Solara."
        solara.Meta(name="description", property="og:description", content=description)
        solara.Meta(name="twitter:description", content=description)
        tags = [
            "jupyter",
            "jupyter dashboard",
            "dashboard",
            "web app",
            "deploy",
            "solara",
        ]
        solara.Meta(name="keywords", content=", ".join(tags))
        with solara.Column():
            Notebook(
                Path(HERE / "_jupyter_dashboard_1.ipynb"),
                show_last_expressions=True,
                execute=False,
                outputs={
                    "a7d17a84": None,  # empty output (7)
                    # original: https://github.com/widgetti/solara/assets/1765949/e844acdb-c77d-4df4-ba4c-a629f92f18a3
                    "82f1d2f7": solara.Image("https://dxhl76zpt6fap.cloudfront.net/pages/docs/content/60-jupyter-dashboard-part1/map.webp"),  # map (11)
                    "3e7ea361": None,  # (13)
                    # original: https://github.com/widgetti/solara/assets/1765949/daaa3a46-61f5-431f-8003-b42b5915da4b
                    "56055643": solara.Image("https://dxhl76zpt6fap.cloudfront.net/pages/docs/content/60-jupyter-dashboard-part1/view.webp"),  # View (15)
                    # original: https://github.com/widgetti/solara/assets/1765949/2f4daf0f-b7d8-4f70-b04a-c27542cffdb0
                    "c78010ec": solara.Image("https://dxhl76zpt6fap.cloudfront.net/pages/docs/content/60-jupyter-dashboard-part1/page.webp"),  # Page (20)
                    # original: https://github.com/widgetti/solara/assets/1765949/a691d9f1-f07b-4e06-b21b-20980476ad64
                    "18290364": solara.Image("https://dxhl76zpt6fap.cloudfront.net/pages/docs/content/60-jupyter-dashboard-part1/controls.webp"),  # Controls
                    "0ca68fe8": None,
                    "fef5d187": None,
                    # original: https://github.com/widgetti/solara/assets/1765949/f0075ad1-808d-458c-8797-e460ce4dc06d
                    "af686391": solara.Image("https://dxhl76zpt6fap.cloudfront.net/pages/docs/content/60-jupyter-dashboard-part1/full-app.webp"),  # Full app
                },
            )
            solara.Markdown(
                """
                Explore this app live at [solara.dev](/apps/jupyter-dashboard-1).

                Don’t miss the next tutorial and stay updated with the latest techniques and insights by subscribing to our newsletter.
            """
            )
            location = solara.use_router().path
            MailChimp(location=location)



"""
# reactive

"""

import solara
from solara.website.components import NoPage
from solara.website.utils import apidoc

title = "reactive"


Page = NoPage


__doc__ += apidoc(solara.reactive)  # type: ignore



import pytest

import solara.server.app
import solara.server.kernel_context
from solara.server import kernel
from solara.server.kernel_context import VirtualKernelContext


@pytest.fixture(autouse=True)
def kernel_context():
    kernel_shared = kernel.Kernel()
    context = VirtualKernelContext(id="1", kernel=kernel_shared, session_id="session-1")
    try:
        with context:
            yield context
    finally:
        with context:
            context.close()


@pytest.fixture()
def no_kernel_context(kernel_context):
    context = solara.server.kernel_context.get_current_context()
    solara.server.kernel_context.set_current_context(None)
    try:
        yield
    finally:
        solara.server.kernel_context.set_current_context(context)



import solara
import ipyvuetify as v


def test_docs_no_browser_simple():
    clicks = solara.reactive(0)

    @solara.component
    def ClickButton():
        def increment():
            clicks.value += 1

        solara.Button(label=f"Clicked: {clicks}", on_click=increment)

    # rc is short for render context
    box, rc = solara.render(ClickButton(), handle_error=False)
    button = box.children[0]
    assert isinstance(button, v.Btn)
    assert button.children[0] == "Clicked: 0"
    # trigger the click event handler without a browser
    button.click()
    assert clicks.value == 1
    assert button.children[0] == "Clicked: 1"



"""
# Sliders

To support proper typechecks, we have multiple slider (all wrapping the ipyvuetify sliders).


"""

import solara
from solara.website.components import NoPage
from solara.website.utils import apidoc

Page = NoPage

__doc__ += "# SliderInt"
__doc__ += apidoc(solara.SliderInt.f)  # type: ignore
__doc__ += "# SliderRangeInt"
__doc__ += apidoc(solara.SliderRangeInt.f)  # type: ignore

__doc__ += "# SliderFloat"
__doc__ += apidoc(solara.SliderFloat.f)  # type: ignore
__doc__ += "# SliderRangeFloat"
__doc__ += apidoc(solara.SliderRangeFloat.f)  # type: ignore

__doc__ += "# SliderValue"
__doc__ += apidoc(solara.SliderValue.f)  # type: ignore

__doc__ += "# SliderDate"
__doc__ += apidoc(solara.SliderDate.f)  # type: ignore



"""# FileBrowser"""

from pathlib import Path
from typing import Optional, cast
import random
import solara
from solara.website.utils import apidoc

opened = solara.reactive(cast(Optional[Path], None))
selected = solara.reactive(cast(Optional[Path], None))
directory = solara.reactive(Path("~").expanduser())
can_select = solara.reactive(False)


@solara.component
def Page():
    def reset_path():
        opened.value = None
        selected.value = None

    def select_random_file():
        files = list(directory.value.glob("*"))
        if files:
            selected.value = random.choice(files)

    # reset path and file when can_select changes
    solara.use_memo(reset_path, [can_select.value])
    solara.Checkbox(label="Enable select", value=can_select)
    solara.FileBrowser(directory, selected=selected, on_file_open=opened.set, can_select=can_select.value)
    solara.Info(f"You are in directory: {directory}")
    solara.Info(f"You selected path: {selected}")
    solara.Info(f"You opened file: {opened}")

    if can_select.value:
        solara.Button(label="Select random file", on_click=select_random_file)


__doc__ += apidoc(solara.FileBrowser.f)  # type: ignore



"""# Interactive sine wave


This example shows how to have two slider control a visualization.

"""

import numpy as np
import plotly.express as px

import solara

x = np.linspace(0, 2, 100)

title = "Interactive sine wave"
freq = solara.reactive(2.0)
phase = solara.reactive(0.1)


@solara.component
def Page():
    y = np.sin(x * freq.value + phase.value)

    solara.FloatSlider("Frequency", value=freq, min=0, max=10)
    solara.FloatSlider("Phase", value=phase, min=0, max=np.pi, step=0.1)

    fig = px.line(x=x, y=y)
    solara.FigurePlotly(fig)



import dataclasses
import threading
import unittest.mock
from pathlib import Path
from typing import Callable, Dict, List, Optional, Set, TypeVar, cast

import pytest

import ipyvuetify as v
import react_ipywidgets as react
from typing_extensions import TypedDict

import solara
import solara.toestand
import solara as sol
import solara.lab
import solara.toestand as toestand
import solara.server.kernel_context
from solara.server import kernel, kernel_context
from solara.toestand import Reactive, Ref, State, use_sync_external_store
import solara.settings
import pandas as pd

from .common import click

HERE = Path(__file__).parent


@dataclasses.dataclass(frozen=True)
class Bears:
    type: str
    count: int = dataclasses.field()


B = TypeVar("B", bound=Bears)


class BearReactive(Reactive[B]):
    def increase_population(self):
        self.update(count=self.get().count + 1)


bears: Bears = Bears(type="brown", count=1)


def get_storage(store):
    storage = store._storage
    if hasattr(storage, "_storage"):
        # this happens when mutation detection is used
        storage = storage._storage
    return storage


def test_store_bare():
    # no need for subclasses
    mock = unittest.mock.Mock()
    mock_change = unittest.mock.Mock()
    store = Reactive[dict]({"string": "foo", "int": 42})
    unsub = store.subscribe(mock)
    unsub_change = store.subscribe_change(mock_change)
    mock.assert_not_called()
    mock_change.assert_not_called()
    store.update(string="bar")
    mock.assert_called_with({"string": "bar", "int": 42})
    mock_change.assert_called_with({"string": "bar", "int": 42}, {"string": "foo", "int": 42})

    unsub()
    unsub_change()


def test_list_index_error():
    mock = unittest.mock.Mock()
    mock_change = unittest.mock.Mock()
    values = solara.reactive([1, 2, 3])
    item3 = Ref(values.fields[2])
    unsub = item3.subscribe(mock)
    unsub_change = item3.subscribe_change(mock_change)
    values.value = [1, 2, 4]
    mock.assert_called_with(4)
    mock_change.assert_called_with(4, 3)
    mock.reset_mock()
    mock_change.reset_mock()
    values.value = [1, 2]
    mock.assert_not_called()
    mock_change.assert_not_called()
    unsub()
    unsub_change()


def test_dict_index_error():
    mock = unittest.mock.Mock()
    mock_change = unittest.mock.Mock()
    values = solara.reactive({"a": 1, "b": 2, "c": 3})
    item3 = Ref(values.fields["c"])
    unsub = item3.subscribe(mock)
    unsub_change = item3.subscribe_change(mock_change)
    values.value = {"a": 1, "b": 2, "c": 4}
    mock.assert_called_with(4)
    mock_change.assert_called_with(4, 3)
    mock.reset_mock()
    mock_change.reset_mock()
    values.value = {"a": 1, "b": 2}
    mock.assert_not_called()
    mock_change.assert_not_called()
    unsub()
    unsub_change()


def test_subscribe():
    bear_store = BearReactive(bears)
    mock = unittest.mock.Mock()
    mock_type = unittest.mock.Mock()
    mock_count = unittest.mock.Mock()
    unsub = []
    unsub += [bear_store.subscribe(mock)]
    unsub += [Ref(bear_store.fields.type).subscribe(mock_type)]
    unsub += [Ref(bear_store.fields.count).subscribe(mock_count)]
    mock.assert_not_called()
    bear_store.update(type="purple")
    mock.assert_called_with(Bears(type="purple", count=1))
    mock_type.assert_called_with("purple")
    mock_count.assert_not_called()

    mock_type.reset_mock()
    bear_store.update(count=3)
    mock.assert_called_with(Bears(type="purple", count=3))
    mock_type.assert_not_called()
    mock_count.assert_called_with(3)
    for u in unsub:
        u()


def test_scopes(no_kernel_context):
    bear_store = BearReactive(bears)
    mock_global = unittest.mock.Mock()
    unsub = []
    unsub += [bear_store.subscribe(mock_global)]

    kernel_shared = kernel.Kernel()
    assert kernel_context.current_context[kernel_context.get_current_thread_key()] is None

    context1 = kernel_context.VirtualKernelContext(id="toestand-1", kernel=kernel_shared, session_id="session-1")
    context2 = kernel_context.VirtualKernelContext(id="toestand-2", kernel=kernel_shared, session_id="session-2")

    with context1:
        mock1 = unittest.mock.Mock()
        unsub += [bear_store.subscribe(mock1)]
    with context2:
        mock2 = unittest.mock.Mock()
        unsub += [bear_store.subscribe(mock2)]

    mock_global.assert_not_called()

    with context2:
        bear_store.update(type="purple")
    mock_global.assert_not_called()
    mock1.assert_not_called()
    mock2.assert_called_with(Bears(type="purple", count=1))
    mock2.reset_mock()

    with context1:
        bear_store.update(count=3)
    mock_global.assert_not_called()
    mock1.assert_called_with(Bears(type="brown", count=3))
    mock2.assert_not_called()
    mock1.reset_mock()

    bear_store.update(type="yellow")
    mock_global.assert_called_with(Bears(type="yellow", count=1))
    mock1.assert_not_called()
    mock2.assert_not_called()

    mock_global.reset_mock()
    for u in unsub:
        u()

    # make sure the use_reactive can be used outside of a kernel context
    reactives = [None, None]
    index = 0

    @solara.component
    def Page():
        reactive = solara.use_reactive(0)
        assert solara.server.kernel_context.get_current_context() is expected_context
        if reactives[index] is None:
            reactives[index] = reactive  # type: ignore
        solara.Info(str(reactive.value))

    with context1:
        expected_context = context1
        _, rc1 = react.render(Page(), handle_error=False)
        assert rc1.find(v.Alert).widget.children[0] == "0"
        assert reactives[0] is not None

    with context2:
        index = 1
        expected_context = context2
        _, rc2 = react.render(Page(), handle_error=False)
        assert rc2.find(v.Alert).widget.children[0] == "0"
        assert reactives[1] is not None

    expected_context = context1
    reactives[0].value = 1
    assert rc1.find(v.Alert).widget.children[0] == "1"
    assert rc2.find(v.Alert).widget.children[0] == "0"

    expected_context = context2
    reactives[1].value = 2
    assert rc1.find(v.Alert).widget.children[0] == "1"
    assert rc2.find(v.Alert).widget.children[0] == "2"

    assert reactives[0].value == 1
    assert reactives[1].value == 2
    with context1:
        assert reactives[0].value == 1
        assert reactives[1].value == 2

    with context1:
        rc1.close()
    with context2:
        rc2.close()


def test_scopes_restore(no_kernel_context):
    kernel1 = kernel.Kernel()
    kernel2 = kernel.Kernel()
    assert kernel_context.current_context[kernel_context.get_current_thread_key()] is None

    context1 = kernel_context.VirtualKernelContext(id="toestand-1", kernel=kernel1, session_id="session-1")
    context2 = kernel_context.VirtualKernelContext(id="toestand-2", kernel=kernel2, session_id="session-2")

    from solara._stores import SharedStore

    reactive: solara.Reactive[int] = solara.Reactive(SharedStore(10))

    @solara.component
    def Page1():
        assert solara.server.kernel_context.get_current_context() is context1
        solara.Info(str(reactive.value))
        Page1b()

    @solara.component
    def Page1b():
        assert solara.server.kernel_context.get_current_context() is context1
        solara.Error(str(reactive.value))

    @solara.component
    def Page2():
        assert solara.server.kernel_context.get_current_context() is context2
        solara.Info(str(reactive.value))

    with context2:
        _, rc2 = react.render(Page2(), handle_error=False)
        assert rc2.find(v.Alert).widget.children[0] == "10"

    with context1:
        _, rc1 = react.render(Page1(), handle_error=False)
        assert rc1.find(v.Alert)[0].widget.children[0] == "10"

    reactive.value = 42
    assert rc1.find(v.Alert)[0].widget.children[0] == "42"
    assert rc2.find(v.Alert).widget.children[0] == "42"

    with context1:
        rc1.close()
    with context2:
        rc2.close()


def test_nested_update():
    # this effectively test the RLock vs Lock
    bear_store = BearReactive(bears)

    mock = unittest.mock.Mock()
    mock_type = unittest.mock.Mock()
    mock_count = unittest.mock.Mock()
    unsub = []
    unsub += [bear_store.subscribe(mock)]
    unsub += [Ref(bear_store.fields.type).subscribe(mock_type)]
    unsub += [Ref(bear_store.fields.count).subscribe(mock_count)]

    def reset_count(new_type):
        bear_store.update(count=0)

    Ref(bear_store.fields.type).subscribe(reset_count)
    Ref(bear_store.fields.type).value = "purple"
    mock.assert_any_call(Bears(type="purple", count=0))
    mock_type.assert_called_with("purple")
    mock_count.assert_called_with(0)
    for u in unsub:
        u()


def test_store_nested():
    @dataclasses.dataclass(frozen=True)
    class City:
        name: str
        population: int

    @dataclasses.dataclass(frozen=True)
    class Person:
        name: str
        height: float

    @dataclasses.dataclass(frozen=True)
    class Country:
        dicator: Person
        name: str
        cities: List[City] = dataclasses.field(default_factory=list)
        people: Dict[str, Person] = dataclasses.field(default_factory=dict)

    # no need for subclasses
    mock = unittest.mock.Mock()
    people = {"Jos": Person(name="Jos", height=1.8)}
    people_copy = people.copy()
    nl = Country(
        name="Netherlands",
        cities=[City(name="Amsterdam", population=1000000)],
        people=people,
        dicator=Person(name="Jos", height=1.8),
    )
    store = Reactive[Country](nl)
    unsub = store.subscribe(mock)
    mock.assert_not_called()
    country = store.fields
    population_accessor = Ref(store.fields.cities[0].population)
    assert population_accessor.get() == 1000000
    population_accessor.set(10)
    assert population_accessor.get() == 10

    mock.assert_called_with(
        Country(
            name="Netherlands",
            cities=[City(name="Amsterdam", population=10)],
            people=people_copy,
            dicator=Person(name="Jos", height=1.8),
        )
    )

    Ref(country.people["Jos"].height).set(1.9)
    assert people == people_copy
    mock.assert_called_with(
        Country(
            name="Netherlands",
            cities=[City(name="Amsterdam", population=10)],
            people={"Jos": Person(name="Jos", height=1.9)},
            dicator=Person(name="Jos", height=1.8),
        )
    )

    Ref(country.people["Jos"].height).value = 2.1
    assert people == people_copy
    mock.assert_called_with(
        Country(
            name="Netherlands",
            cities=[City(name="Amsterdam", population=10)],
            people={"Jos": Person(name="Jos", height=2.1)},
            dicator=Person(name="Jos", height=1.8),
        )
    )

    # TODO: new name?
    # Ref(country.people["Jos"].height).set(lambda x: x + 0.1)
    # mock.assert_called_with(
    #     Country(
    #         name="Netherlands",
    #         cities=[City(name="Amsterdam", population=10)],
    #         people={"Jos": Person(name="Jos", height=2.0)},
    #         dicator=Person(name="Jos", height=1.8),
    #     ),
    # )

    Ref(country.people).update(lambda x: {**x, "Maria": Person(name="Maria", height=1.7)})
    mock.assert_called_with(
        Country(
            dicator=Person(name="Jos", height=1.8),
            name="Netherlands",
            cities=[City(name="Amsterdam", population=10)],
            people={
                "Jos": Person(name="Jos", height=2.1),
                "Maria": Person(name="Maria", height=1.7),
            },
        )
    )
    Ref(country.dicator.height).set(2.0)
    mock.assert_called_with(
        Country(
            dicator=Person(name="Jos", height=2.0),
            name="Netherlands",
            cities=[City(name="Amsterdam", population=10)],
            people={
                "Jos": Person(name="Jos", height=2.1),
                "Maria": Person(name="Maria", height=1.7),
            },
        )
    )

    unsub()


def test_bear_store_basics():
    mock = unittest.mock.Mock()
    bear_store = BearReactive(bears)
    unsub = bear_store.subscribe(mock)
    mock.assert_not_called()
    bear_store.increase_population()
    mock.assert_called_with(Bears(type="brown", count=2))
    bear_store.increase_population()
    assert mock.call_count == 2
    mock.assert_called_with(Bears(type="brown", count=3))

    setter = bear_store.setter(bear_store.fields.count)
    setter(5)
    assert mock.call_count == 3
    mock.assert_called_with(Bears(type="brown", count=5))

    unsub()
    bear_store.increase_population()
    assert mock.call_count == 3

    # now test a subfield
    mock_count = unittest.mock.Mock()
    count = Ref(bear_store.fields.count)
    bear_store.subscribe(mock)
    count.subscribe(mock_count)
    count.set(4)
    assert count.get() == 4
    assert bear_store.get() == Bears(type="brown", count=4)
    mock.assert_called_with(Bears(type="brown", count=4))
    mock_count.assert_called_with(4)


def test_bear_store_basics_dict():
    class Bears(TypedDict):
        type: str
        count: int

    bears = Bears(type="brown", count=1)

    class BearReactive(Reactive[Bears]):
        def increase_population(self):
            self.update(count=self.get()["count"] + 1)

    mock = unittest.mock.Mock()
    bear_store = BearReactive(bears)
    unsub = bear_store.subscribe(mock)
    mock.assert_not_called()
    bear_store.increase_population()
    mock.assert_called_with(Bears(type="brown", count=2))
    bear_store.increase_population()
    assert mock.call_count == 2
    mock.assert_called_with(Bears(type="brown", count=3))

    setter = bear_store.setter(bear_store.fields["count"])  # type: ignore
    setter(5)  # type: ignore
    assert mock.call_count == 3
    mock.assert_called_with(Bears(type="brown", count=5))

    unsub()
    bear_store.increase_population()
    assert mock.call_count == 3


def test_bear_store_react():
    bear_store = BearReactive(bears)

    @react.component
    def BearCounter():
        bear_count = bear_store.use(lambda bear: bear.count)
        return sol.Info(f"{bear_count} bears around here")

    @react.component
    def Controls():
        return sol.Button("add bear", on_click=bear_store.increase_population)

    @react.component
    def App():
        with sol.VBox() as main:
            BearCounter()
            Controls()
        return main

    el = App()
    box, rc = react.render(el, handle_error=False)
    assert rc._find(v.Alert).widget.children[0] == "1 bears around here"
    click(rc._find(v.Btn).widget)
    assert rc._find(v.Alert).widget.children[0] == "2 bears around here"

    # the storage should live in the app context to support multiple users/connections
    kernel_shared = kernel.Kernel()
    context1 = kernel_context.VirtualKernelContext(id="bear-1", kernel=kernel_shared, session_id="session-1")
    context2 = kernel_context.VirtualKernelContext(id="bear-2", kernel=kernel_shared, session_id="session-2")
    rcs = []
    for context in [context1, context2]:
        with context:
            el = App()
            box, rc = react.render(el, handle_error=False)
            rcs.append(rc)
            # el = App()
            # box, rc = react.render(el, handle_error=False)
            assert rc._find(v.Alert).widget.children[0] == "1 bears around here"
            assert Ref(bear_store.fields.count).value == 1
            click(rc._find(v.Btn).widget)
            assert rc._find(v.Alert).widget.children[0] == "2 bears around here"
            assert Ref(bear_store.fields.count).value == 2

    rc = rcs[1]
    with context2:
        rc.render(el)
        assert Ref(bear_store.fields.count).value == 2
        assert rc._find(v.Alert).widget.children[0] == "2 bears around here"
        click(rc._find(v.Btn).widget)
        assert Ref(bear_store.fields.count).value == 3
        assert rc._find(v.Alert).widget.children[0] == "3 bears around here"

    rc = rcs[0]
    with context1:
        rc.render(el)
        assert Ref(bear_store.fields.count).value == 2
        assert rc._find(v.Alert).widget.children[0] == "2 bears around here"
        click(rc._find(v.Btn).widget)
        assert Ref(bear_store.fields.count).value == 3
        assert rc._find(v.Alert).widget.children[0] == "3 bears around here"
        click(rc._find(v.Btn).widget)
        assert Ref(bear_store.fields.count).value == 4
        assert rc._find(v.Alert).widget.children[0] == "4 bears around here"
        click(rc._find(v.Btn).widget)
        assert Ref(bear_store.fields.count).value == 5
        assert rc._find(v.Alert).widget.children[0] == "5 bears around here"

    rc = rcs[1]
    with context2:
        rc.render(el)
        assert Ref(bear_store.fields.count).value == 3
        assert rc._find(v.Alert).widget.children[0] == "3 bears around here"
        click(rc._find(v.Btn).widget)
        assert Ref(bear_store.fields.count).value == 4
        assert rc._find(v.Alert).widget.children[0] == "4 bears around here"
        Ref(bear_store.fields.count).value = 10
        assert rc._find(v.Alert).widget.children[0] == "10 bears around here"


def test_simplest():
    settings = State({"bears": 2, "theme": "dark"})
    unsub = settings.subscribe(print)  # noqa

    settings.update(theme="light")
    # prints: {"bears": 2, "theme": "light"}

    unsub()  # remove event listener
    theme_accessor = Ref(settings.fields["theme"])  # type: ignore

    # Now use it in a React component

    import react_ipywidgets as react

    renders = 0

    import solara as sol

    @react.component
    def ThemeInfo():
        # the lambda function here is called a selector, it 'selects' out the state you want
        # theme = settings.use(lambda state: state["theme"])
        theme = Ref(settings.fields["theme"]).use_value()  # type: ignore
        return sol.Info(f"Using theme {theme}")

    @react.component
    def ThemeSelector():
        theme, set_theme = Ref(settings.fields["theme"]).use_state()  # type: ignore
        with sol.ToggleButtonsSingle(theme, on_value=set_theme) as main:  # type: ignore
            sol.Button("dark")
            sol.Button("light")
        return main

    @react.component
    def Test():
        nonlocal renders
        renders += 1
        with sol.VBox() as main:
            ThemeInfo()
            ThemeSelector()
        return main

    box, rc = react.render(Test(), handle_error=False)
    assert rc._find(v.Alert).widget.children[0] == "Using theme light"
    assert rc._find(v.BtnToggle).widget.v_model == 1
    theme_accessor.set("dark")
    assert rc._find(v.Alert).widget.children[0] == "Using theme dark"
    assert rc._find(v.BtnToggle).widget.v_model == 0
    rc._find(v.BtnToggle).widget.v_model = 1
    assert rc._find(v.Alert).widget.children[0] == "Using theme light"
    renders_before = renders
    settings.update(bears=3)
    assert renders == renders_before


@dataclasses.dataclass(frozen=True)
class Fish:
    fishes: int


F = TypeVar("F", bound=Fish)


class FishReactive(Reactive[F]):
    def jump(self):
        print("jump")  # noqa


@dataclasses.dataclass(frozen=True)
class AppStateComposite:
    bear: Bears
    fish: Fish


class AppState(Reactive[AppStateComposite]):
    bears: BearReactive

    def __post__init__(self):
        # for composite stores, manually create the substores
        self.bears = BearReactive(Ref(self.fields.bear))
        # it's ok not to have a substore for fish

    def eat(self):
        state = self.get()
        bears = state.bear.count
        fish = Fish(fishes=max(0, state.fish.fishes - bears))
        self.update(fish=fish)


def test_store_primitive():
    string_value = State("foo")
    mock = unittest.mock.Mock()
    unsub = string_value.subscribe(mock)
    mock.assert_not_called()
    string_value.set("bar")
    mock.assert_called_with("bar")
    unsub()


def test_store_computed():
    list_store = Reactive[list]([1, 2, 3])

    count = list_store.computed(len)
    last = list_store.computed(lambda x: x[-1] if x else None)

    assert count._auto_subscriber.value.reactive_used is None
    assert count.get() == 3
    assert count._auto_subscriber.value.reactive_used == {list_store}

    assert last._auto_subscriber.value.reactive_used is None
    assert last.get() == 3
    assert last._auto_subscriber.value.reactive_used == {list_store}
    mock = unittest.mock.Mock()
    mock_last = unittest.mock.Mock()
    unsub = count.subscribe(mock)
    unsub_last = last.subscribe(mock_last)
    mock.assert_not_called()
    mock_last.assert_not_called()
    list_store.set([1, 2, 3, 42])
    mock.assert_called_with(4)
    mock_last.assert_called_with(42)
    unsub()
    unsub_last()
    list_store.set([])
    mock.assert_called_with(4)
    mock_last.assert_called_with(42)


def test_app_composite():
    mock = unittest.mock.Mock()
    fish_state = Fish(fishes=4)
    app_state = AppStateComposite(bear=bears, fish=fish_state)
    app_store = AppState(app_state)
    app_store.subscribe(mock)

    assert app_store.get().bear.count == 1
    assert app_store.bears.get().count == 1
    assert app_store.get().fish.fishes == 4

    mock.assert_not_called()
    app_store.eat()
    assert app_store.get().fish.fishes == 3
    assert mock.call_count == 1
    mock.assert_called_with(AppStateComposite(bear=Bears(type="brown", count=1), fish=Fish(fishes=3)))

    app_store.bears.increase_population()  # type: ignore
    mock.assert_called_with(AppStateComposite(bear=Bears(type="brown", count=2), fish=Fish(fishes=3)))
    assert mock.call_count == 2
    assert app_store.get().bear.count == 2
    assert app_store.bears.get().count == 2
    assert app_store.get().fish.fishes == 3

    app_store.eat()
    assert app_store.get().fish.fishes == 1
    assert mock.call_count == 3
    mock.assert_called_with(AppStateComposite(bear=Bears(type="brown", count=2), fish=Fish(fishes=1)))
    app_store.eat()
    assert app_store.get().fish.fishes == 0
    assert mock.call_count == 4
    mock.assert_called_with(AppStateComposite(bear=Bears(type="brown", count=2), fish=Fish(fishes=0)))


@dataclasses.dataclass(frozen=True)
class AppInherit(Bears, Fish):
    pass


class AppStateInherit(BearReactive[AppInherit], FishReactive[AppInherit], Reactive[AppInherit]):
    def eat(self):
        state = self.get()
        bears = state.count
        self.update(fishes=max(0, state.fishes - bears))


def test_app_inherit():
    mock = unittest.mock.Mock()
    app = AppInherit(type="brown", count=1, fishes=4)
    app_store = AppStateInherit(app)
    app_store.subscribe(mock)

    assert app_store.get().count == 1
    assert app_store.get().fishes == 4

    mock.assert_not_called()
    app_store.eat()
    assert app_store.get().fishes == 3
    assert mock.call_count == 1
    mock.assert_called_with(AppInherit(type="brown", count=1, fishes=3))

    app_store.increase_population()
    mock.assert_called_with(AppInherit(type="brown", count=2, fishes=3))
    assert mock.call_count == 2
    assert app_store.get().count == 2
    assert app_store.get().count == 2
    assert app_store.get().fishes == 3

    app_store.eat()
    assert app_store.get().fishes == 1
    assert mock.call_count == 3
    mock.assert_called_with(AppInherit(type="brown", count=2, fishes=1))
    app_store.eat()
    assert app_store.get().fishes == 0
    assert mock.call_count == 4
    mock.assert_called_with(AppInherit(type="brown", count=2, fishes=0))


def test_use_external_store():
    state = {"a": 1, "b": 2, "c": 3}
    update = lambda: None  # noqa

    def subscribe(callback: Callable[[], None]) -> Callable[[], None]:
        nonlocal update
        update = callback

        def cleanup():
            nonlocal update
            update = lambda: None  # noqa

        return cleanup

    renders = 0
    state_last = {}

    @react.component
    def Test():
        nonlocal renders
        nonlocal state_last
        state_last = use_sync_external_store(subscribe, lambda: state)
        renders += 1
        return sol.Button()

    box, rc = react.render(Test(), handle_error=False)
    assert renders == 1
    update()
    assert renders == 1
    assert state_last == state
    state = state.copy()
    state["b"] = 12
    update()
    assert state_last == state
    assert renders == 2


# TODO:
# def test_store_scope_application():
#     import solara as sol
#     import solara.scope

#     bear_store = BearReactive({"type": "brown", "count": 1}, storage=sol.scope.application)
#     try:
#         assert bear_store.get() == {"type": "brown", "count": 1}

#         @react.component
#         def BearCounter():
#             bears = bear_store.use(lambda bear: bear.count)
#             # bears = app_store.use(lambda app: app.bear.count)
#             return sol.Info(f"{bears} bears around here")

#         @react.component
#         def Controls():
#             return sol.Button("add bear", on_click=bear_store.increase_population)

#         @react.component
#         def App():
#             with sol.VBox() as main:
#                 BearCounter()
#                 Controls()
#             return main

#         app = App()
#         box, rc = react.render(app, handle_error=False)

#         assert rc._find(v.Alert).widget.children[0] == "1 bears around here"
#         bear_store.increase_population()
#         assert rc._find(v.Alert).widget.children[0] == "2 bears around here"


#         storage_key = bear_store.storage_key
#         redis_dict = bear_store._storage.observable_dict
#         redis_dict[storage_key] = {"type": "brown", "count": 3}
#         # TODO: wait for the store to update
#         time.sleep(0.1)
#         # assert rc._find(v.Alert).wait_for_widget.children[0] == "3 bears around here"
#         # assert rc._find(v.Alert).wait_for_single().widget.children[0] == "3 bears around here"
#         rc._find(v.Alert).widget.children[0] == "3 bears around here"
#     finally:
#         bear_store._storage.delete()
def test_dataframe():
    mock = unittest.mock.Mock()
    import pandas as pd

    df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    df2 = df.copy()
    df2["a"] = [1, 2, 4]
    store = Reactive[pd.DataFrame](df)
    unsub = store.subscribe(mock)
    mock.assert_not_called()
    store.set(df)
    mock.assert_not_called()
    store.set(df2)
    mock.assert_called_with(df2)

    unsub()
    store.set(df)

    @solara.component
    def Test():
        df = store.use_value()
        return solara.Info(repr(id(df)))

    box, rc = solara.render(Test(), handle_error=False)

    if solara.settings.storage.mutation_detection:
        # a copy is made, so get a reference to the actual used object
        df = get_storage(store).value.public
    assert rc.find(v.Alert).widget.children[0] == repr(id(df))
    df2 = df2.copy()
    store.set(df2)
    if solara.settings.storage.mutation_detection:
        df2 = get_storage(store).value.public
    assert rc.find(v.Alert).widget.children[0] == repr(id(df2))

    @dataclasses.dataclass(frozen=True)
    class DataFrame:
        df: Optional[pd.DataFrame] = dataclasses.field()

    df_store = Reactive[DataFrame](DataFrame(df=df))
    mock.reset_mock()
    unsub = Ref(df_store.fields).subscribe(mock)
    mock.assert_not_called()
    df_store.set(DataFrame(df=df2))
    mock.assert_called_once()
    mock.reset_mock()
    Ref(df_store.fields.df).set(df)
    mock.assert_called_once()
    unsub()


def test_thread_local():
    def test1():
        assert solara.lab.thread_local.reactive_used is None

    t = threading.Thread(target=test1)
    t.start()
    t.join()

    def test2():
        myset: Set[solara.lab.ValueBase] = set()
        solara.lab.local.reactive_used = myset
        assert solara.lab.thread_local.reactive_used is myset

    t = threading.Thread(target=test2)
    t.start()
    t.join()

    def test3():
        assert solara.lab.thread_local.reactive_used is None

    t = threading.Thread(target=test3)
    t.start()
    t.join()


def test_reactive_auto_subscribe(kernel_context):
    x = Reactive(1)
    y = Reactive("hi")
    extra = Reactive("extra")
    count = Reactive(1)

    @solara.component
    def Test():
        if x.value == 0:
            _ = extra.value  # access conditional
        _ = x.value  # access twice
        return solara.IntSlider(label=y.value, value=x.value)

    @solara.component
    def Main():
        for i in range(count.value):
            Test()
        if count.value == 0:
            return solara.Info("no slider")

    box, rc = solara.render(Main(), handle_error=False)
    assert rc.find(v.Slider).widget.v_model == 1
    x.value = 2
    assert rc.find(v.Slider).widget.v_model == 2
    y.value = "hello"
    assert rc.find(v.Slider).widget.label == "hello"
    assert len(get_storage(x).listeners2) == 1
    # force an extra listener
    x.value = 0
    # and remove it
    x.value = 1

    count.value = 2
    assert len(rc.find(v.Slider)) == 2
    assert len(get_storage(x).listeners2[kernel_context.id]) == 2
    x.value = 3
    assert rc.find(v.Slider)[0].widget.v_model == 3
    assert len(get_storage(x).listeners2[kernel_context.id]) == 2

    count.value = 1
    assert len(rc.find(v.Slider)) == 1
    count.value = 0
    assert len(rc.find(v.Slider)) == 0

    rc.close()
    assert not get_storage(x).listeners[kernel_context.id]
    assert not get_storage(x).listeners2[kernel_context.id]


def test_reactive_auto_subscribe_sub():
    bears = Reactive(Bears(type="brown", count=1))
    renders = 0

    ref = Ref(bears.fields.count)
    reactive_used = None

    @solara.component
    def Test():
        nonlocal reactive_used
        nonlocal renders
        reactive_used = toestand.thread_local.reactive_used
        renders += 1
        count = ref.value
        return solara.Info(f"{count} bears around here")

    box, rc = solara.render(Test(), handle_error=False)
    assert rc.find(v.Alert).widget.children[0] == "1 bears around here"
    assert reactive_used == {ref}
    ref.value += 1
    assert rc.find(v.Alert).widget.children[0] == "2 bears around here"
    assert reactive_used == {ref}
    # now check that we didn't listen to the whole object, just count changes
    renders_before = renders
    Ref(bears.fields.type).value = "pink"
    assert renders == renders_before


def test_reactive_auto_subscribe_cleanup(kernel_context):
    x = Reactive(1)
    y = Reactive("hi")
    renders = 0

    @solara.component
    def Test():
        nonlocal renders
        renders += 1
        if x.value == 42:
            _ = y.value  # access conditional
        if x.value == 0:
            _ = y.value  # access conditional
            x.value = 100
        return solara.IntSlider("test", value=x.value)

    box, rc = solara.render(Test(), handle_error=False)
    assert rc.find(v.Slider).widget.v_model == 1
    assert renders == 1
    assert len(get_storage(x).listeners2) == 1
    assert len(get_storage(y).listeners2) == 0
    x.value = 42
    assert renders == 2
    assert len(get_storage(x).listeners2[kernel_context.id]) == 1
    assert len(get_storage(y).listeners2[kernel_context.id]) == 1

    # this triggers two renders, where during the first one we use y, but the seconds we don't
    x.value = 0
    assert rc.find(v.Slider).widget.v_model == 100
    assert len(get_storage(x).listeners2[kernel_context.id]) == 1
    # which means we shouldn't have a listener on y
    assert len(get_storage(y).listeners2[kernel_context.id]) == 0

    rc.close()
    assert not get_storage(x).listeners[kernel_context.id]
    assert not get_storage(y).listeners2[kernel_context.id]


def test_reactive_auto_subscribe_subfield_limit(kernel_context):
    bears = Reactive(Bears(type="brown", count=1))
    renders = 0
    reactive_used = None

    @solara.component
    def Test():
        nonlocal renders
        nonlocal reactive_used
        reactive_used = toestand.thread_local.reactive_used
        renders += 1
        _ = bears.value  # access it to trigger the subscription
        return solara.IntSlider("test", value=Ref(bears.fields.count).value)

    box, rc = solara.render(Test(), handle_error=False)
    assert rc.find(v.Slider).widget.v_model == 1
    assert renders == 1
    assert reactive_used is not None
    assert len(reactive_used) == 2  # bears and bears.fields.count
    Ref(bears.fields.count).value = 2
    assert renders == 2
    assert len(reactive_used) == 2  # bears and bears.fields.count
    rc.close()
    assert not bears._storage.listeners[kernel_context.id]
    assert not bears._storage.listeners2[kernel_context.id]


def test_reactive_batch_update():
    count = Reactive(1)
    mock1 = unittest.mock.Mock()
    mock2 = unittest.mock.Mock()

    @solara.component
    def Test1():
        mock1(count.value)
        return solara.IntSlider("test", value=count.value)

    @solara.component
    def Test2():
        mock2(count.value)
        return solara.IntSlider("test", value=count.value)

    @solara.component
    def Test():
        Test1()
        Test2()

    box, rc = solara.render(Test(), handle_error=False)
    assert rc.find(v.Slider)[0].widget.v_model == 1
    assert rc.find(v.Slider)[1].widget.v_model == 1
    assert rc.render_count == 1
    assert mock1.call_count == 1
    assert mock2.call_count == 1
    count.value = 2
    assert mock1.call_count == 2
    assert mock2.call_count == 2
    assert rc.find(v.Slider)[0].widget.v_model == 2
    assert rc.find(v.Slider)[1].widget.v_model == 2
    assert mock1.call_count == 2
    assert mock2.call_count == 2
    assert rc.render_count == 2


def test_repr():
    x = Reactive(1)
    assert repr(x).startswith("<Reactive value=1")
    assert str(x) == "1"
    y = Reactive("hi")
    assert repr(y).startswith("<Reactive value='hi'")
    assert str(y) == "'hi'"

    class Foo:
        bar = Reactive(1)

    assert repr(Foo.bar).startswith("<Reactive Foo.bar value=1")
    assert str(Foo.bar) == "Foo.bar=1"

    bears = Reactive(Bears(type="brown", count=1))
    s = repr(bears.fields.count)
    assert s.startswith("<Field <Reactive value=Bears(type='brown', count=1)")
    assert s.endswith(".count>")


def test_use_reactive_update():
    control = Reactive(0)
    var1 = Reactive(1)
    var2 = Reactive(2)

    @solara.component
    def Test():
        var: Reactive[int]
        if control.value == 0:
            var = solara.use_reactive(var1)
        else:
            var = solara.use_reactive(var2)

        return solara.IntSlider("test: " + str(var.value), value=var)

    box, rc = solara.render(Test(), handle_error=False)
    assert rc.find(v.Slider).widget.v_model == 1
    assert rc.find(v.Slider).widget.label == "test: 1"
    control.value = 1
    assert rc.find(v.Slider).widget.v_model == 2
    assert rc.find(v.Slider).widget.label == "test: 2"
    # the slider should be using the same reactive variable (var2)
    rc.find(v.Slider).widget.v_model = 1
    assert var2.value == 1
    assert rc.find(v.Slider).widget.label == "test: 1"
    rc.close()


def test_use_reactive_ref():
    reactive_var = Reactive({"a": 1})
    reactive_ref = Ref(reactive_var.fields["a"])

    reactive_ref_test: Optional[Reactive[int]] = None

    @solara.component
    def Test():
        nonlocal reactive_ref_test
        reactive_ref_test = solara.use_reactive(reactive_ref)
        return solara.IntSlider("test: " + str(reactive_ref.value), value=reactive_ref_test)

    box, rc = solara.render(Test(), handle_error=False)
    assert reactive_ref_test is not None
    assert reactive_ref_test.value == 1
    rc.close()


def test_use_reactive_on_change():
    control = Reactive(0)
    var1 = Reactive(1)
    var2 = Reactive(2)
    mock1 = unittest.mock.Mock()
    mock2 = unittest.mock.Mock()

    @solara.component
    def Test():
        var: Reactive[int]
        if control.value == 0:
            var = solara.use_reactive(var1, on_change=mock1)
        else:
            var = solara.use_reactive(var2, on_change=mock2)

        return solara.IntSlider("test", value=var)

    box, rc = solara.render(Test(), handle_error=False)
    assert rc.find(v.Slider).widget.v_model == 1
    assert mock1.call_count == 0
    assert mock2.call_count == 0
    # if it changes downstream, it should trigger on_change
    rc.find(v.Slider).widget.v_model = 10
    assert mock1.call_count == 1
    assert mock2.call_count == 0

    control.value = 1
    assert mock1.call_count == 1
    assert mock2.call_count == 0
    assert rc.find(v.Slider).widget.v_model == 2
    rc.find(v.Slider).widget.v_model = 20
    assert mock1.call_count == 1
    assert mock2.call_count == 1

    control.value = 0
    assert rc.find(v.Slider).widget.v_model == 10
    rc.find(v.Slider).widget.v_model = 100
    assert mock1.call_count == 2
    assert mock2.call_count == 1

    rc.close()


def test_reactive_var_in_use_effect():
    var = Reactive(1)

    @solara.component
    def Test():
        def modify():
            var.value = 2

        solara.use_effect(modify)
        # note: just pass the value, not the reactive variable
        # otherwise the test passes. It is important *this*
        # component listens to the reactive variable.
        return solara.IntSlider("test", value=var.value)

    box, rc = solara.render(Test(), handle_error=False)
    assert rc.find(v.Slider).widget.v_model == 2


def test_singleton():
    from solara.toestand import Singleton

    calls = 0

    def factory():
        nonlocal calls
        calls += 1
        return Bears(type="brown", count=1)

    s = Singleton(factory)
    assert calls == 0
    assert s.get() == Bears(type="brown", count=1)
    assert calls == 1
    assert s.get() == Bears(type="brown", count=1)
    assert calls == 1


def test_computed():
    context_id = "1"
    from solara.toestand import Computed

    x = Reactive(1)
    y = Reactive(2)
    calls = 0

    def conditional_add():
        nonlocal calls
        calls += 1
        if x.value == 0:
            return 42
        else:
            return x.value + y.value

    z = Computed(conditional_add)
    assert z._auto_subscriber.value.reactive_used is None
    assert z.value == 3
    assert z._auto_subscriber.value.reactive_used == {x, y}
    # assert z._auto_subscriber.subscribed == 1
    assert len(get_storage(x).listeners[context_id]) == 0
    assert len(get_storage(x).listeners2[context_id]) == 1
    assert len(get_storage(y).listeners[context_id]) == 0
    assert len(get_storage(y).listeners2[context_id]) == 1
    assert calls == 1
    x.value = 2
    assert z.value == 4
    assert z._auto_subscriber.value.reactive_used == {x, y}
    assert calls == 2
    y.value = 3
    assert z.value == 5
    assert z._auto_subscriber.value.reactive_used == {x, y}
    assert calls == 3
    assert len(get_storage(x).listeners2[context_id]) == 1
    assert len(get_storage(y).listeners2[context_id]) == 1

    # now we do not depend on y anymore
    x.value = 0
    assert z.value == 42
    assert z._auto_subscriber.value.reactive_used == {x}
    assert len(get_storage(x).listeners2[context_id]) == 1
    assert len(get_storage(y).listeners2[context_id]) == 0
    assert calls == 4
    y.value = 4
    assert z.value == 42
    assert z._auto_subscriber.value.reactive_used == {x}
    assert calls == 4


def test_computed_reload(no_kernel_context):
    import solara.server.reload
    from solara.server.app import AppScript

    name = str(HERE / "toestand_computed_reload.py")
    # if we reload, the _type counter will reset, but we will not
    # execute all reactive variables again, which causes a mismatch between
    # the reactive variable id's
    solara.toestand.KernelStore._type_counter.clear()
    app = AppScript(name)
    app.init()
    try:
        assert len(app.routes) == 1
        route = app.routes[0]
        c = app.run()
        kernel_shared = kernel.Kernel()
        kernel_context = solara.server.kernel_context.VirtualKernelContext(id="1", kernel=kernel_shared, session_id="session-1")
        with kernel_context:
            root = solara.RoutingProvider(children=[c], routes=app.routes, pathname="/")
            box, rc = solara.render(root, handle_error=False)
            kernel_context.app_object = rc
            text = rc.find(v.TextField)
            assert text.widget.v_model == "2.0"
            route.module.value_reactive.value = 2  # type: ignore
            module = route.module
            assert text.widget.v_model == "3.0"
            solara.server.reload.reloader.requires_reload = True
            assert solara.server.reload.reloader.on_change is not None
            solara.server.reload.reloader.on_change("notimportant.py")
            kernel_context.restart()
        solara.toestand.KernelStore._type_counter.clear()
        c = app.run()
        with kernel_context:
            route = app.routes[0]
            root = solara.RoutingProvider(children=[c], routes=app.routes, pathname="/")
            box, rc = solara.render(root, handle_error=False)
            kernel_context.app_object = rc
            text = rc.find(v.TextField)
            assert text.widget.v_model == "3.0"
            assert route.module is not module
            route.module.value_reactive.value = 3  # type: ignore
            assert text.widget.v_model == "4.0"
    finally:
        app.close()


def test_pydantic_basic():
    from pydantic import BaseModel

    class Person(BaseModel):
        name: str
        height: float

    person = Reactive[Person](Person(name="Jos", height=1.8))
    assert person.get() == Person(name="Jos", height=1.8)
    assert person.get().name == "Jos"
    assert person.get().height == 1.8
    assert Ref(person.fields.name).get() == "Jos"
    assert Ref(person.fields.height).get() == 1.8

    person.set(Person(name="Maria", height=1.7))
    assert person.get() == Person(name="Maria", height=1.7)
    assert person.get().name == "Maria"
    assert person.get().height == 1.7
    assert Ref(person.fields.name).get() == "Maria"
    assert Ref(person.fields.height).get() == 1.7

    Ref(person.fields.height).set(2.0)
    assert person.get() == Person(name="Maria", height=2.0)
    assert person.get().name == "Maria"
    assert person.get().height == 2.0
    assert Ref(person.fields.name).get() == "Maria"
    assert Ref(person.fields.height).get() == 2.0


@pytest.fixture
def mutation_detection():
    current = solara.settings.storage.mutation_detection
    solara.settings.storage.mutation_detection = True
    try:
        yield
    finally:
        solara.settings.storage.mutation_detection = current


def test_mutate_initial_value(no_kernel_context, mutation_detection):
    initial_values: List[int] = [1, 2, 3]
    reactive = Reactive[List[int]](solara.toestand.mutation_detection_storage(initial_values))
    assert reactive.value == initial_values
    initial_values.append(4)
    with pytest.raises(ValueError, match="Reactive variable was initialized"):
        reactive.value = [3, 2, 1]
    kernel_shared = kernel.Kernel()
    context = kernel_context.VirtualKernelContext(id="toestand-1", kernel=kernel_shared, session_id="session-1")
    assert kernel_context.current_context[kernel_context.get_current_thread_key()] is None
    with pytest.raises(ValueError, match="Reactive variable was initialized"):
        with context:
            # reading the initial values should also trigger the error
            _ = reactive.value


@pytest.mark.skipif(not solara.settings.storage.mutation_detection, reason="only tests when SOLARA_STORAGE_MUTATION_DETECTION=1")
def test_mutate_initial_value_local(no_kernel_context, mutation_detection):
    initial_values: List[int] = [1, 2, 3]
    reactive_local = solara.reactive([4, 5, 6])

    @solara.component
    def Page():
        nonlocal reactive_local
        reactive_local = solara.use_reactive(initial_values)

    _, rc = react.render(Page(), handle_error=False)
    try:
        assert reactive_local.value == initial_values
        initial_values.append(4)
        with pytest.raises(ValueError, match="Reactive variable was initialized"):
            reactive_local.value = [3, 2, 1]
        kernel_shared = kernel.Kernel()
        context = kernel_context.VirtualKernelContext(id="toestand-1", kernel=kernel_shared, session_id="session-1")
        assert kernel_context.current_context[kernel_context.get_current_thread_key()] is None
        with pytest.raises(ValueError, match="Reactive variable was initialized"):
            with context:
                # reading the initial values should also trigger the error
                _ = reactive_local.value
    finally:
        rc.close()


def test_mutate_value_public_value():
    values = [1, 2, 3]

    reactive = Reactive[List[int]](solara.toestand.mutation_detection_storage(values))
    reactive.value.append(4)
    with pytest.raises(ValueError, match="Reactive variable was read when it had the value of \\[1, 2, 3\\].*"):
        reactive._storage.check_mutations()  # type: ignore


@pytest.mark.skipif(not solara.settings.storage.mutation_detection, reason="only tests when SOLARA_STORAGE_MUTATION_DETECTION=1")
def test_mutate_value_public_value_local():
    values = [1, 2, 3]
    reactive = solara.reactive(values)

    @solara.component
    def Page():
        nonlocal reactive
        reactive = solara.use_reactive(values)

    _, rc = react.render(Page(), handle_error=False)
    try:
        reactive.value.append(4)
        with pytest.raises(ValueError, match="Reactive variable was read when it had the value of \\[1, 2, 3\\].*"):
            _ = reactive.value
    finally:
        rc.close()


def test_mutate_value_set_value():
    values = [1, 2, 3]
    reactive = Reactive[List[int]](solara.toestand.mutation_detection_storage(values))
    new_values = [1, 2, 3, 4]
    reactive.value = new_values
    new_values.append(5)
    # print(reactive.value)
    with pytest.raises(ValueError, match="Reactive variable was set.*"):
        reactive._storage.check_mutations()  # type: ignore


def test_mutate_value_set_value_dataframe():
    # dataframes do not support simple equality checks
    # and we cannot have false equals(a, b) == False result when they
    # actually are equal
    df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    df_orig = df.copy()
    reactive_df = Reactive[Optional[pd.DataFrame]](
        solara.toestand.mutation_detection_storage(cast(Optional[pd.DataFrame], None), equals=solara.util.equals_pickle)
    )
    assert reactive_df._storage.equals is solara.util.equals_pickle
    assert reactive_df._storage.equals(df, df_orig)
    reactive_df.value = df
    df["a"][0] = 100
    assert not reactive_df._storage.equals(df, df_orig)
    with pytest.raises(ValueError, match="Reactive variable was set.*"):
        reactive_df._storage.check_mutations()  # type: ignore



import solara
import ipyvuetify as v


def test_docs_no_browser_api_find():
    clicks = solara.reactive(0)

    @solara.component
    def ClickButton():
        def increment():
            clicks.value += 1

        with solara.Card("Button in a card"):
            with solara.Column().meta(ref="somecolumn"):
                solara.Button(label=f"Clicked: {clicks}", on_click=increment)
            with solara.Column():
                solara.Button(label="Not the button we need")

    # rc is short for render context
    box, rc = solara.render(ClickButton(), handle_error=False)
    # this find will make the .widget fail, because it matches two buttons
    # finder = rc.find(v.Btn)
    # We can refine our search by adding constraints to attributes of the widget
    button_locator = rc.find(v.Btn, children=["Clicked: 0"])
    # basics asserts are supported, like assert_single(), assert_empty(), assert_not_empty()
    button_locator.assert_single()
    button = button_locator.widget
    # .find calls can also be nested, and can use the meta_ref to find the right widget
    # finder = rc.find(meta_ref="somecolumn").find(v.Btn)
    button.click()
    assert clicks.value == 1
    rc.find(v.Btn, children=["Clicked: 1"]).assert_single()



"""
# Avatar

"""

from types import ModuleType
from typing import Optional

auth: Optional[ModuleType]
try:
    from solara_enterprise import auth
except ImportError:
    auth = None
from solara.website.components import NoPage
from solara.website.utils import apidoc

title = "Avatar"

Page = NoPage

if auth:
    __doc__ += apidoc(auth.Avatar.f)  # type: ignore
else:
    __doc__ += "solara-enterprise not installed."



from pathlib import Path

from solara.website.components.markdown import MarkdownWithMetadata
from solara.website.components.sidebar import Sidebar


title = "Roadmap"
HERE = Path(__file__)
Sidebar = Sidebar

Page = MarkdownWithMetadata(Path(HERE.parent / "roadmap.md").read_text())



from typing import Dict, List, Optional, Union

import reacton.ipyvuetify as v
import solara
from solara.util import _combine_classes


@solara.component
def Card(
    title: Optional[str] = None,
    subtitle: Optional[str] = None,
    elevation: int = 2,
    margin=2,
    children: List[solara.Element] = [],
    classes: List[str] = [],
    style: Union[str, Dict[str, str], None] = None,
):
    """A card combines a title, subtitle, content and actions into a single unit.


    ## Example
    ```solara
    import solara

    @solara.component
    def Page():
        with solara.Card(title="Card title", subtitle="Card subtitle"):
            solara.Markdown(
                "Lorem ipsum dolor sit amet consectetur adipisicing elit. "\\
                "Commodi, ratione debitis quis est labore voluptatibus! "\\
                "Eaque cupiditate minima, at placeat totam, magni doloremque "\\
                "veniam neque porro libero rerum unde voluptatem!"
            )
            with solara.CardActions():
                solara.Button("Action 1", text=True)
                solara.Button("Action 2", text=True)
    ```


    ## Arguments

     * `title`: Title of the card.
     * `subtitle`: Subtitle of the card.
     * `elevation`: Elevation of the card, gives the appearance of hovering above the page.
     * `margin`: Margin of the card.
     * `children`: Children are placed as the main content of the card.
     * `style`: CSS style to apply to the top level element.
    """
    class_ = _combine_classes([f"ma-{margin}", *classes])
    style_flat = solara.util._flatten_style(style)
    children_actions = []
    children_text = []
    for child in children:
        if isinstance(child, solara.Element) and child.component == solara.CardActions:
            children_actions.extend(child.kwargs.get("children", []))
        else:
            children_text.append(child)
    with v.Card(elevation=elevation, class_=class_, style_=style_flat) as main:
        if title:
            with v.CardTitle(
                children=[title],
            ):
                pass
        if subtitle:
            with v.CardSubtitle(
                children=[subtitle],
            ):
                pass
        with v.CardText(children=children_text):
            pass
        if children_actions:
            with v.CardActions(children=children_actions):
                pass
    return main


@solara.component
def CardActions(children: List[solara.Element] = []):
    """Container for actions in a card.

    See [Card](/documentation/components/layout/card) for an example.

    # Arguments

     * `children`: Children are placed as the action area of the card.
    """
    solara.Error("You should not see this if you add a CardActions as a child component of a Card")



from typing import Callable, Dict, List, Optional, Union

import solara
import solara.util
from reacton import ipyvue
from reacton import ipyvuetify as v


@solara.component
def Button(
    label: str = None,
    on_click: Callable[[], None] = None,
    icon_name: str = None,
    children: list = [],
    disabled=False,
    text=False,
    outlined=False,
    color: Optional[str] = None,
    click_event="click",
    classes: List[str] = [],
    style: Union[str, Dict[str, str], None] = None,
    value=None,
    **kwargs,
):
    """A button that can be clicked to trigger an event.

    ## Example

    ```solara
    import solara


    @solara.component
    def Page():
        with solara.Row():
            solara.Button(label="Default")
            solara.Button(label="Default+color", color="primary")
            solara.Button(label="Text", text=True)
            solara.Button(label="Outlined", outlined=True)
            solara.Button(label="Outlined+color", outlined=True, color="primary")
    ```


    ## Arguments

    - `label`: The text to display on the button.
    - `on_click`: A callback function that is called when the button is clicked.
    - `icon_name`: The name of the icon to display on the button ([Overview of available icons](https://pictogrammers.github.io/@mdi/font/4.9.95/)).
    - `children`: A list of child elements to display on the button.
    - `disabled`: Whether the button is disabled.
    - `text`: Whether the button should be displayed as text, it has no shadow and no background.
    - `outlined`: Whether the button should be displayed as outlined, it has no background.
    - `value`: (Optional) When used as a child of a ToggleButtons component, the value of the selected button, see
        [ToggleButtons](/documentation/components/input/togglebuttons).
    - `classes`: Additional CSS classes to apply.
    - `style`: CSS style to apply.

    ### Deprecated arguments
    - click_event: (Deprecated/export option: The event that triggers the on_click callback, which can include vue event modifiers).

    """
    if label:
        children = [label] + children
    if icon_name:
        children = [v.Icon(left=bool(label), children=[icon_name])] + children
    if "class_" in kwargs:
        kwargs = kwargs.copy()
        class_ = solara.util._combine_classes([*classes, kwargs.pop("class_")])
    else:
        class_ = solara.util._combine_classes(classes)
    kwargs = kwargs.copy()
    style_flat = solara.util._flatten_style(style)
    if "style_" in kwargs:
        style_flat += kwargs.pop("style_")
    if solara.util.ipyvuetify_major_version == 3:
        variant = "elevated"
        if text:
            variant = "text"
        elif outlined:
            variant = "outlined"
        btn = solara.v.Btn(children=children, **kwargs, disabled=disabled, class_=class_, style_=style_flat, color=color, variant=variant)
    else:
        btn = solara.v.Btn(children=children, **kwargs, disabled=disabled, text=text, class_=class_, style_=style_flat, outlined=outlined, color=color)
    ipyvue.use_event(btn, click_event, lambda *_ignore: on_click and on_click())
    return btn



from unittest.mock import MagicMock

import ipyvuetify as vw

import solara


def test_select():
    """
    test select widget
    """
    disabled = None
    on_value = MagicMock()
    on_value_multiple = MagicMock()

    @solara.component
    def Test():
        nonlocal disabled
        disabled = solara.use_reactive(False)
        solara.Select(label="single", values=["test0", "test1", "test2"], on_value=on_value, disabled=disabled.value)
        solara.SelectMultiple(label="multiple", values=[], all_values=["test0", "test1", "test2"], on_value=on_value_multiple, disabled=disabled.value)

    _, rc = solara.render(Test(), handle_error=False)
    select = rc.find(vw.Select, label="single").widget
    select_multi = rc.find(vw.Select, label="multiple").widget

    # init
    assert not select.disabled
    assert not select_multi.disabled

    assert select.v_model is None
    assert select_multi.v_model == []

    # set v_model
    select.v_model = "test0"
    assert on_value.call_count == 1
    assert on_value.call_args[0][0] == "test0"

    select_multi.v_model = ["test0", "test1"]
    assert on_value_multiple.call_count == 1
    assert on_value_multiple.call_args[0][0] == ["test0", "test1"]

    # test disable
    assert disabled is not None
    disabled.set(True)
    assert len(rc.find(vw.Select, disabled=True).widgets) == 2

    rc.close()



"""
abstract base class for websocket with a sync interface.
Async implementation have to come up with a way how to do this sync (see e.g. the starlette implementation)
"""

import abc
import json
from typing import Union


class WebSocketDisconnect(Exception):
    pass


class WebsocketWrapper(abc.ABC):
    @abc.abstractmethod
    def send_text(self, data: str) -> None:
        pass

    @abc.abstractmethod
    def send_bytes(self, data: bytes) -> None:
        pass

    def send(self, data: Union[str, bytes]) -> None:
        if isinstance(data, str):
            self.send_text(data)
        elif isinstance(data, bytes):
            self.send_bytes(data)
        else:
            raise TypeError(f"{type(data)} not supported, only ")

    def send_json(self, data):
        self.send_text(json.dumps(data))

    @abc.abstractmethod
    def close(self) -> None:
        pass

    @abc.abstractmethod
    async def receive(self) -> Union[str, bytes]:
        pass

    async def receive_json(self):
        text = await self.receive()
        return json.loads(text)



from collections.abc import Hashable
import hashlib
import inspect
import logging
import sys
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    MutableMapping,
    Optional,
    TypeVar,
    Union,
    cast,
    overload,
)

try:
    import cachetools

    has_cachetools = True
except ModuleNotFoundError:
    has_cachetools = False

import solara
import solara.settings
import solara.util
import typing_extensions
from reacton.utils import equals

logger = logging.getLogger("solara.cache")

# placeholder value for missing keys
_DOES_NOT_EXIST = object()

_global_values_used: Dict[Any, Dict[str, Any]] = {}


T = TypeVar("T")
R = TypeVar("R")
K = TypeVar("K")
V = TypeVar("V")
P = typing_extensions.ParamSpec("P")

Storage = MutableMapping[K, V]


if has_cachetools:

    class Memory(cachetools.LRUCache):
        def __init__(self, max_items=solara.settings.cache.memory_max_items):
            super().__init__(maxsize=max_items)

else:

    class Memory(dict):  # type: ignore
        pass


def _default_key(*args, **kwargs):
    kwargs_tuple: Any = ()
    for key, value in kwargs.items():
        kwargs_tuple += (key, value)
    return (args, kwargs_tuple)


class MemoizedFunction(Generic[P, R]):
    def __init__(self, function: Callable[P, R], key: Callable[P, Hashable], storage: Optional[Storage], allow_nonlocals=False):
        self.function = function
        f: Callable = self.function
        if not allow_nonlocals:
            nonlocals = inspect.getclosurevars(f).nonlocals
            if nonlocals:
                raise ValueError(f"Memoized functions cannot depend on nonlocal variables, it now depends on {nonlocals}")
        if sys.version_info[:2] < (3, 9):
            # usedforsecurity is only available in Python 3.9+
            codehash = hashlib.md5(f.__code__.co_code).hexdigest()
        else:
            codehash = hashlib.md5(f.__code__.co_code, usedforsecurity=False).hexdigest()  # type: ignore

        self.function_key = (f.__qualname__, codehash)
        current_globals = dict(inspect.getclosurevars(f).globals)
        _global_values_used.setdefault(self.function_key, current_globals)
        self._check_globals()
        self.key = key
        self._storage = storage
        self.intrusive_cancel = True
        # modifications to these are not thread safe
        # but good enough for an indication
        self.hits = 0
        self.misses = 0

    @property
    def storage(self) -> Storage:
        return self._storage if self._storage is not None else storage

    def _check_globals(self):
        globals = _global_values_used.get(self.function_key)
        current_globals = dict(inspect.getclosurevars(self.function).globals)
        if current_globals is not globals:
            if not equals(current_globals, globals):
                raise ValueError(
                    f"Memoized functions depend on globals variables, but they changed. The first value was {globals}, but now it is {current_globals}"
                )

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        self._check_globals()
        key = (self.function_key, self.key(*args, **kwargs))
        # if .get is atomic, we do not need to lock
        value = cast(R, self.storage.get(key, _DOES_NOT_EXIST))
        if value is _DOES_NOT_EXIST:
            self.misses += 1
            value = self.function(*args, **kwargs)
            self.storage[key] = value
        else:
            self.hits += 1
        return value

    def use_thread(self, *args: P.args, **kwargs: P.kwargs) -> solara.Result[R]:
        """Calls the function in a thread when not in cache, otherwise returns the cached value.

        This is almost similar to a use_thread on the memoized function, except that
        we ensure that if the function return value is already in the cache, we don't
        have to wait for the thread to finish.

        This can cause less flickering in the UI, since it will immediately return a
        'FINISHED' result, instead of a 'RUNNING' result when possible.
        """
        key = (self.function_key, self.key(*args, **kwargs))

        # we assume generating the key is cheap, but a storage.get might not be
        # since it may have to ask the a server, so we wrap it in use_memo
        def get_current():
            return cast(R, self.storage.get(key, _DOES_NOT_EXIST))

        value = solara.use_memo(get_current, dependencies=[key])

        def do_work():
            if value is _DOES_NOT_EXIST:
                self.misses += 1
                self._check_globals()
                new_value = self.function(*args, **kwargs)
                # TODO: what if someone else already put in the value?
                # we do a check in vaex if it differs
                # but we could also store a Future
                self.storage[key] = new_value
                return new_value
            else:
                # although we don't use the return value directly, it's still used on the next time result_thread is
                # returned.
                return value

        result_thread: solara.Result[R] = solara.use_thread(do_work, dependencies=[key], intrusive_cancel=self.intrusive_cancel)
        if value is _DOES_NOT_EXIST:
            return result_thread
        else:
            self.hits += 1
            # what are the semantics of cancel and retry here?
            return solara.Result(value, state=solara.ResultState.FINISHED)


@overload
def memoize(
    function: None = None,
    key: None = None,
    storage: Optional[Storage] = None,
    allow_nonlocals=False,
) -> Callable[[Callable[P, R]], MemoizedFunction[P, R]]: ...


@overload
def memoize(
    function: None = None,
    key: Callable[P, Hashable] = ...,
    storage: Optional[Storage] = None,
    allow_nonlocals=False,
) -> Callable[[Callable[P, R]], MemoizedFunction[P, R]]: ...


@overload
def memoize(
    function: Callable[P, R],
    key: None = None,
    storage: None = None,
    allow_nonlocals=False,
) -> MemoizedFunction[P, R]: ...


def memoize(
    function: Union[None, Callable[P, R]] = None,
    key: Union[None, Callable[P, Hashable]] = None,
    storage: Optional[Storage] = None,
    allow_nonlocals: bool = False,
) -> Union[Callable[[Callable[P, R]], MemoizedFunction[P, R]], MemoizedFunction[P, R]]:
    """Will cache function return values based on the arguments.

    Example:

    ```python
    @solara.memoize
    def mean(values):
         return sum(values) / len(values)
    ```

    If a key function is provided, it will be used to generate the cache key based on the arguments instead.
    This is useful in situations where the arguments are not hashable, but a unique key can be generated
    based on the arguments.

    Also used in situations where the arguments are expensive to hash, but a cheaper key
    that is unique exists. For example using the filename instead of the file content.

    Example:

    ```python
    @solara.memoize(key=lambda df, column: (id(df), column))
    def mean(df, column):
        return df[column].mean()
    ```

    Without the key function, the above would fail because a DataFrame is not hashable.

    The function name and function code are added to the argument based key, together making up a unique cache key.
    Nonlocals variables are not included in the key, and are not allowed by default. Pass `allow_nonlocals=True`
    to allow nonlocals variables, or add them as arguments to the function.
    Globals are by default allowed, and also not included in the key. Solara will try to detect if globals are
    changed, and if so, will raise a ValueError.

    If a storage is provided, it will be used to store the cached values. If no storage is provided, the
    shared storage will be used. This is useful in situations where the cache should be shared between
    different functions to avoid excessive memory usage by limiting the number of cache entries or the
    memory content.

    The storage can be any object that implements the MutableMapping interface, for instance a dict or
    a cachetools.LRUCache. Or a new instance of `solara.cache.Memory`, see [caching](/documentation/advanced/reference/caching)
    for cache storage options.

    The return value of the decorator behaves like the original function, but also has a few attributes:

     * `storage`: the storage used to cache the values
     * `key`: the key function used to generate the key
     * `function`: the original function
     * `use_thread`: a hook that will execute the function in a thread, and return a Result object
        If the value is already cached, the function will not be executed in a thread.


    See also the [reference on caching](/documentation/advanced/reference/caching) for more caching details.

    """

    def wrapper(func: Callable[P, R]) -> MemoizedFunction[P, R]:
        return MemoizedFunction[P, R](
            func,
            cast(Callable[P, Hashable], key or _default_key),
            storage,
            allow_nonlocals,
        )

    if function is None:
        return wrapper
    else:
        return wrapper(function)


cache_type_map = {
    "memory": "solara.cache.Memory",
    "memory-size": "solara_enterprise.cache.memory_size.MemorySize",
    "disk": "solara_enterprise.cache.disk.Disk",
    "redis": "solara_enterprise.cache.redis.Redis",
    "multi-level": "solara_enterprise.cache.multi_level.MultiLevel",
}


def create(name: str = "memory", *args, **kwargs) -> Storage:
    """Create a cache storage based on the name.

    ## Arguments

     - `name`: the name of the cache type, can be one of `memory`, `memory-size`, `disk`, `redis`, `multi-level`
     - `*args`: the arguments to pass to the cache constructor
     - `**kwargs`: the keyword arguments to pass to the cache constructor

    """
    if "," in name:
        names = name.split(",")
        logger.info("Set multilevel cache to %r", names)
        caches = []
        for name in names:
            caches.append(create(name, *args, **kwargs))
        cache = create("multi-level", *caches)
    elif name in cache_type_map:
        cls = solara.util.import_item(cache_type_map[name])
        cache = cls(*args, **kwargs)
    else:
        raise ValueError(f"Unknown type of cache {name}")
    return cache


storage: Storage = create(solara.settings.cache.type)


def configure(name="memory", *args, **kwargs):
    """Shorthand for solara.cache.storage = solara.cache.create(name, *args, **kwargs)"""
    global storage
    storage = create(name, *args, **kwargs)



import os

import playwright.sync_api

# from solara.util import cwd


def test_create_button(page_session: playwright.sync_api.Page, solara_server, solara_app, tmpdir):
    script = tmpdir / "button.py"
    assert os.system(f"solara create button {script}") == 0
    with solara_app(str(script)):
        page_session.goto(solara_server.base_url)
        page_session.locator("text=Clicked: 0").click()
        page_session.locator("text=Clicked: 1").click()


def test_create_markdown(page_session: playwright.sync_api.Page, solara_server, solara_app, tmpdir):
    script = tmpdir / "button.py"
    assert os.system(f"solara create markdown {script}") == 0
    with solara_app(str(script)):
        page_session.goto(solara_server.base_url)
        page_session.locator("text=Renders like").wait_for()


# def test_create_portal(page_session: playwright.sync_api.Page, solara_server, solara_app, tmpdir):
#     package_dir = tmpdir / "solara-portal-pytest"
#     assert os.system(f"solara create portal {package_dir}") == 0
#     with cwd(package_dir):
#         assert os.system("pip install .") == 0
#     with solara_app("solara_portal_pytest.pages"):
#         page_session.goto(solara_server.base_url)
#         page_session.locator("text=Scatter for titanic").wait_for()
#         page_session.locator("text=Open table view").first.click()
#         page_session.locator("text=pclass").wait_for()
#         page_session.go_back()
#         page_session.locator("text=Read article").first.click()
#         page_session.locator("text=Substiterat vati").first.wait_for()
#         page_session.go_back()
#         page_session.locator("text=Scatter for titanic").wait_for()
#         page_session.wait_for_timeout(100)
#         page_session.locator("text=Scatter for titanic").click()
#         page_session.locator("text=x").first.wait_for()



import copy
import dataclasses
import inspect
import sys
import threading
from typing import Callable, ContextManager, Generic, Optional, Union, cast, Any
import warnings


from .toestand import ValueBase, S, _find_outside_solara_frame, _DEBUG

import solara.util
import solara.settings


class _PublicValueNotSet:
    pass


class _SetValueNotSet:
    pass


@dataclasses.dataclass
class StoreValue(Generic[S]):
    private: S  # the internal private value, should never be mutated
    public: Union[S, _PublicValueNotSet]  # this is the value that is exposed in .get(), it is a deep copy of private
    get_traceback: Optional[inspect.Traceback]
    set_value: Union[S, _SetValueNotSet]  # the value that was set using .set(..), we deepcopy this to set private
    set_traceback: Optional[inspect.Traceback]


class MutateDetectorStore(ValueBase[S]):
    def __init__(self, store: ValueBase[StoreValue[S]], equals=solara.util.equals_extra):
        self._storage = store
        self._enabled = True
        super().__init__(equals=equals)

    @property
    def lock(self):
        return self._storage.lock

    def get(self) -> S:
        self.check_mutations()
        self._ensure_public_exists()
        value = self._storage.get()
        # value.public is of type Optional[S], so it's tempting to check for None here,
        # but S could include None as a valid value, so best we can do is cast
        public_value = cast(S, value.public)
        return public_value

    def peek(self) -> S:
        """Return the value without automatically subscribing to listeners."""
        self.check_mutations()
        store_value = self._storage.peek()
        self._ensure_public_exists()
        public_value = cast(S, store_value.public)
        return public_value

    def set(self, value: S):
        self.check_mutations()
        self._ensure_public_exists()
        private = copy.deepcopy(value)
        self._check_equals(private, value)
        frame = _find_outside_solara_frame()
        if frame is not None:
            frame_info = inspect.getframeinfo(frame)
        else:
            frame_info = None
        store_value = StoreValue(private=private, public=_PublicValueNotSet(), get_traceback=None, set_value=value, set_traceback=frame_info)
        self._storage.set(store_value)

    def check_mutations(self):
        self._storage._check_mutation()
        if not self._enabled:
            return
        store_value = self._storage.peek()
        if not isinstance(store_value.public, _PublicValueNotSet) and not self.equals(store_value.public, store_value.private):
            tb = store_value.get_traceback
            # TODO: make the error message as elaborate as below
            msg = (
                f"Reactive variable was read when it had the value of {store_value.private!r}, but was later mutated to {store_value.public!r}.\n"
                "Mutation should not be done on the value of a reactive variable, as in production mode we would be unable to track changes.\n"
            )
            if tb:
                if tb.code_context:
                    code = tb.code_context[0]
                else:
                    code = "<No code context available>"
                msg += f"The last value was read in the following code:\n{tb.filename}:{tb.lineno}\n{code}"
            raise ValueError(msg)
        elif not isinstance(store_value.set_value, _SetValueNotSet) and not self.equals(store_value.set_value, store_value.private):
            tb = store_value.set_traceback
            msg = f"""Reactive variable was set with a value of {store_value.private!r}, but was later mutated mutated to {store_value.set_value!r}.

Mutation should not be done on the value of a reactive variable, as in production mode we would be unable to track changes.

Bad:
    mylist = reactive([]]
    some_values = [1, 2, 3]
    mylist.value = some_values  # you give solara a reference to your list
    some_values.append(4)  # but later mutate it (solara cannot detect this change, so a render will not be triggered)
    # if later on a re-render happens for a different reason, you will read of the mutated list.

Good (if you want the reactive variable to be updated):
    mylist = reactive([]]
    some_values = [1, 2, 3]
    mylist.value = some_values
    mylist.value = some_values + [4]

Good (if you want to keep mutating your own list):
    mylist = reactive([]]
    some_values = [1, 2, 3]
    mylist.value = some_values.copy()  # this gives solara a copy of the list
    some_values.append(4)  # you are free to mutate your own list, solara will not see this

"""
            if tb:
                if tb.code_context:
                    code = tb.code_context[0]
                else:
                    code = "<No code context available>"
                msg += f"The last time the value was set was at:\n{tb.filename}:{tb.lineno}\n{code}"
            raise ValueError(msg)

    def _ensure_public_exists(self):
        store_value = self._storage.peek()
        if isinstance(store_value.public, _PublicValueNotSet):
            with self.lock:
                if isinstance(store_value.public, _PublicValueNotSet):
                    frame = _find_outside_solara_frame()
                    if frame is not None:
                        frame_info = inspect.getframeinfo(frame)
                    else:
                        frame_info = None
                    store_value.public = copy.deepcopy(store_value.private)
                    self._check_equals(store_value.public, store_value.private)
                    store_value.get_traceback = frame_info

    def _check_equals(self, a: S, b: S):
        if not self._enabled:
            return
        if not self.equals(a, b):
            frame = _find_outside_solara_frame()
            if frame is not None:
                frame_info = inspect.getframeinfo(frame)
            else:
                frame_info = None

            warn = """The equals function for this reactive value returned False when comparing a deepcopy to itself.

This reactive variable will not be able to detect mutations correctly, and is therefore disabled.

To avoid this warning, and to ensure that mutation detection works correctly, please provide a better equals function to the reactive variable.
A good choice for dataframes and numpy arrays might be solara.util.equals_pickle, which will also attempt to compare the pickled values of the objects.

Example:
df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
reactive_df = solara.reactive(df, equals=solara.util.equals_pickle)
"""
            tb = frame_info
            if tb:
                if tb.code_context:
                    code = tb.code_context[0]
                else:
                    code = "<No code context available>"
                warn += f"This warning was triggered from:\n{tb.filename}:{tb.lineno}\n{code}"
            warnings.warn(warn)
            self._enabled = False

    def subscribe(self, listener: Callable[[S], None], scope: Optional[ContextManager] = None):
        def listener_wrapper(new: StoreValue[S], previous: StoreValue[S]):
            self._ensure_public_exists()
            assert not isinstance(new.public, _PublicValueNotSet)
            assert not isinstance(previous.public, _PublicValueNotSet)
            previous_value = previous.set_value if not isinstance(previous.set_value, _SetValueNotSet) else previous.private
            new_value = new.set_value
            assert not isinstance(new_value, _SetValueNotSet)
            if not self.equals(new_value, previous_value):
                listener(new_value)

        return self._storage.subscribe_change(listener_wrapper, scope=scope)

    def subscribe_change(self, listener: Callable[[S, S], None], scope: Optional[ContextManager] = None):
        def listener_wrapper(new: StoreValue[S], previous: StoreValue[S]):
            self._ensure_public_exists()
            assert not isinstance(new.public, _PublicValueNotSet)
            assert not isinstance(previous.public, _PublicValueNotSet)
            previous_value = previous.set_value if not isinstance(previous.set_value, _SetValueNotSet) else previous.private
            new_value = new.set_value
            assert not isinstance(new_value, _SetValueNotSet)
            if not self.equals(new_value, previous_value):
                listener(new_value, previous_value)

        return self._storage.subscribe_change(listener_wrapper, scope=scope)


class SharedStore(ValueBase[S]):
    """Stores a single value, not kernel scoped."""

    _traceback: Optional[inspect.Traceback]
    _original_ref: Optional[S]
    _original_ref_copy: Optional[S]

    def __init__(self, value: S, equals: Callable[[Any, Any], bool] = solara.util.equals_extra, unwrap=lambda x: x):
        # since a set can trigger events, which can trigger new updates, we need a recursive lock
        self._lock = threading.RLock()
        self.local = threading.local()
        self.equals = equals

        self._value = value
        self._original_ref = None
        self._original_ref_copy = None
        self._unwrap = unwrap
        self._mutation_detection = solara.settings.storage.mutation_detection
        if self._mutation_detection:
            frame = _find_outside_solara_frame()
            if frame is not None:
                self._traceback = inspect.getframeinfo(frame)
            else:
                self._traceback = None
            self._original_ref = value
            self._original_ref_copy = copy.deepcopy(self._original_ref)
            if not self.equals(self._unwrap(self._original_ref), self._unwrap(self._original_ref_copy)):
                msg = """The equals function for this reactive value returned False when comparing a deepcopy to itself.

This reactive variable will not be able to detect mutations correctly, and is therefore disabled.

To avoid this warning, and to ensure that mutation detection works correctly, please provide a better equals function to the reactive variable.
A good choice for dataframes and numpy arrays might be solara.util.equals_pickle, which will also attempt to compare the pickled values of the objects.

Example:
df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
reactive_df = solara.reactive(df, equals=solara.util.equals_pickle)
"""
                tb = self._traceback
                if tb:
                    if tb.code_context:
                        code = tb.code_context[0]
                    else:
                        code = "<No code context available>"
                    msg += f"This warning was triggered from:\n{tb.filename}:{tb.lineno}\n{code}"
                warnings.warn(msg)
                self._mutation_detection = False
        super().__init__(equals=equals)

    def _check_mutation(self):
        if not self._mutation_detection:
            return
        current = self._unwrap(self._original_ref)
        initial = self._unwrap(self._original_ref_copy)
        if not self.equals(initial, current):
            tb = self._traceback
            if tb:
                if tb.code_context:
                    code = tb.code_context[0].strip()
                else:
                    code = "No code context available"
                msg = f"Reactive variable was initialized at {tb.filename}:{tb.lineno} with {initial!r}, but was mutated to {current!r}.\n{code}"
            else:
                msg = f"Reactive variable was initialized with a value of {initial!r}, but was mutated to {current!r} (unable to report the location in the source code)."
            raise ValueError(msg)

    @property
    def lock(self):
        return self._lock

    def peek(self):
        self._check_mutation()
        return self._value

    def get(self):
        self._check_mutation()
        return self._value

    def clear(self):
        pass

    def _get_scope_key(self):
        return "GLOBAL"

    def set(self, value: S):
        self._check_mutation()
        old = self.get()
        if self.equals(old, value):
            return
        self._value = value

        if _DEBUG:
            import traceback

            traceback.print_stack(limit=17, file=sys.stdout)

            print("change old", old)  # noqa
            print("change new", value)  # noqa

        self.fire(value, old)



import atexit
import io
import sys
from binascii import b2a_base64
from threading import local
from unittest.mock import Mock

import reacton.patch_display
from IPython.core.displaypub import DisplayPublisher
from IPython.core.interactiveshell import InteractiveShell, InteractiveShellABC
from jupyter_client.session import Session, extract_header
from traitlets import Any, CBytes, Dict, Instance, Type, default


def encode_images(obj):
    # no-op in ipykernel
    return obj


def json_clean(obj):
    # no-op in ipykernel
    return obj


# based on the zmq display publisher from ipykernel
# ideally this goes out of ipykernel
class SolaraDisplayPublisher(DisplayPublisher):
    """A display publisher that publishes data using a ZeroMQ PUB socket."""

    session = Instance(Session, allow_none=True)
    pub_socket = Any(allow_none=True)
    parent_header = Dict({})
    topic = CBytes(b"display_data")

    _thread_local = Any()

    def set_parent(self, parent):
        """Set the parent for outbound messages."""
        self.parent_header = extract_header(parent)

    def _flush_streams(self):
        """flush IO Streams prior to display"""
        sys.stdout.flush()
        sys.stderr.flush()

    @default("_thread_local")
    def _default_thread_local(self):
        """Initialize our thread local storage"""
        return local()

    @property
    def _hooks(self):
        if not hasattr(self._thread_local, "hooks"):
            # create new list for a new thread
            self._thread_local.hooks = []
        return self._thread_local.hooks

    def publish(
        self,
        data,
        metadata=None,
        source=None,
        *,  # Enforce keyword-only arguments to match DisplayPublisher.publish
        transient=None,
        update=False,
        **kwargs,  # Make sure we're compatible with DisplayPublisher.publish
    ) -> None:
        """Publish a display-data message

        Parameters
        ----------
        data : dict
            A mime-bundle dict, keyed by mime-type.
        metadata : dict, optional
            Metadata associated with the data.
        transient : dict, optional, keyword-only
            Transient data that may only be relevant during a live display,
            such as display_id.
            Transient data should not be persisted to documents.
        update : bool, optional, keyword-only
            If True, send an update_display_data message instead of display_data.
        """
        self._flush_streams()
        if metadata is None:
            metadata = {}
        if transient is None:
            transient = {}
        self._validate_data(data, metadata)
        content = {}
        content["data"] = encode_images(data)
        content["metadata"] = metadata
        content["transient"] = transient

        msg_type = "update_display_data" if update else "display_data"

        # Use 2-stage process to send a message,
        # in order to put it through the transform
        # hooks before potentially sending.
        assert self.session is not None
        msg = self.session.msg(msg_type, json_clean(content), parent=self.parent_header)

        # Each transform either returns a new
        # message or None. If None is returned,
        # the message has been 'used' and we return.
        for hook in self._hooks:
            msg = hook(msg)
            if msg is None:
                return

        self.session.send(
            self.pub_socket,
            msg,
            ident=self.topic,
        )

    def clear_output(self, wait=False):
        """Clear output associated with the current execution (cell).

        Parameters
        ----------
        wait : bool (default: False)
            If True, the output will not be cleared immediately,
            instead waiting for the next display before clearing.
            This reduces bounce during repeated clear & display loops.

        """
        assert self.session is not None
        content = dict(wait=wait)
        self._flush_streams()
        msg = self.session.msg("clear_output", json_clean(content), parent=self.parent_header)
        for hook in self._hooks:
            msg = hook(msg)
            if msg is None:
                return

        self.session.send(
            self.pub_socket,
            msg,
            ident=self.topic,
        )

    def register_hook(self, hook):
        """
        Registers a hook with the thread-local storage.

        Parameters
        ----------
        hook : Any callable object

        Returns
        -------
        Either a publishable message, or `None`.
        The DisplayHook objects must return a message from
        the __call__ method if they still require the
        `session.send` method to be called after transformation.
        Returning `None` will halt that execution path, and
        session.send will not be called.
        """
        self._hooks.append(hook)

    def unregister_hook(self, hook):
        """
        Un-registers a hook with the thread-local storage.

        Parameters
        ----------
        hook : Any callable object which has previously been
            registered as a hook.

        Returns
        -------
        bool - `True` if the hook was removed, `False` if it wasn't
            found.
        """
        try:
            self._hooks.remove(hook)
            return True
        except ValueError:
            return False


class SolaraInteractiveShell(InteractiveShell):
    display_pub_class = Type(SolaraDisplayPublisher)
    history_manager = Any()  # type: ignore
    display_pub: SolaraDisplayPublisher

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        atexit.unregister(self.atexit_operations)

        if self.magics_manager:
            magic = self.magics_manager.registry["ScriptMagics"]
            atexit.unregister(magic.kill_bg_processes)

    def set_parent(self, parent):
        """Tell the children about the parent message."""
        self.display_pub.set_parent(parent)

    def init_sys_modules(self):
        pass  # don't create a __main__, it will cause a mem leak

    def init_prefilter(self):
        pass  # avoid consuming memory

    def init_history(self):
        self.history_manager = Mock()  # type: ignore

    def init_display_formatter(self):
        super().init_display_formatter()
        assert self.display_formatter is not None
        self.display_formatter.ipython_display_formatter = reacton.patch_display.ReactonDisplayFormatter()

        # matplotlib support for display(figure)
        # IPython.core.pylabtools has support for this, but it requires importing matplotlib
        # which would slow down startup, so we do it here using for_type using a string as argument.
        def encode_png(figure, **kwargs):
            f = io.BytesIO()
            format = "png"
            figure.savefig(f, format=format, **kwargs)
            bytes_data = f.getvalue()
            base64_data = b2a_base64(bytes_data, newline=False).decode("ascii")
            return base64_data

        formatter = self.display_formatter.formatters["image/png"]
        formatter.for_type("matplotlib.figure.Figure", encode_png)

    def init_display_pub(self):
        super().init_display_pub()
        self.display_pub.register_hook(self.display_in_reacton_hook)

    def display_in_reacton_hook(self, msg):
        """Will intercept a display call and add the display data to an output widget when in a reacton context/render function."""
        # similar to reacton.patch_display.publish
        from reacton.core import get_render_context

        rc = get_render_context(required=False)
        # only during the render phase we want to capture the display calls
        # during the reconsolidation phase we want to let the original display publisher do its thing
        # such as adding it to a output widget
        if rc is not None and not rc.reconsolidating and msg["msg_type"] == "display_data":
            from reacton.ipywidgets import Output

            Output(
                outputs=[
                    {
                        "output_type": "display_data",
                        "data": msg["content"]["data"],
                        "metadata": msg["content"]["metadata"],
                    }
                ]
            )
            return None  # do not send to the frontend
        return msg

    def reset(self, new_session=True, aggressive=False):
        # if we dont override this with a dummy, the unittests will take a long time
        # to exit. Hitting ctrl-c showed the following stacktrace:
        # Traceback (most recent call last):
        # File "/Users/maartenbreddels/miniconda3/envs/dev/lib/python3.9/site-packages/IPython/core/interactiveshell.py", line 3924, in atexit_operations
        #     self.reset(new_session=False)
        # File "/Users/maartenbreddels/miniconda3/envs/dev/lib/python3.9/site-packages/IPython/core/interactiveshell.py", line 1456, in reset
        #     self.displayhook.flush()
        # File "/Users/maartenbreddels/miniconda3/envs/dev/lib/python3.9/site-packages/IPython/core/displayhook.py", line 311, in flush
        #     gc.collect()
        # So this dummy will prevent that.
        pass


InteractiveShellABC.register(SolaraInteractiveShell)



"""
# GridFixed

Lays out children in a grid with a fixed number of columns.
"""

import solara

from ..common import ColorCard

title = "GridFixed"


@solara.component
def Page():
    colors = "green red orange brown yellow pink".split()
    with solara.GridFixed(columns=3):
        for color in colors:
            ColorCard(color, color)



import asyncio
import logging
import mimetypes
import os
import threading
from http.server import HTTPServer
from pathlib import Path
from typing import Any, Dict, List
from urllib.parse import urlparse
from uuid import uuid4

import flask
import simple_websocket
from flask import Blueprint, Flask, abort, request, send_from_directory, url_for
from flask_sock import Sock

try:
    import solara_enterprise  # type: ignore

    del solara_enterprise

    has_solara_enterprise = True
except ImportError:
    has_solara_enterprise = False

if has_solara_enterprise:
    from solara_enterprise.auth.flask import allowed  # type: ignore
    from solara_enterprise.auth.flask import (
        authorize,
        get_user,
        init_flask,
        login,
        logout,
    )
else:

    def allowed():
        return True


import solara
import solara.settings
from solara.server.threaded import ServerBase

from . import app as appmod
from . import cdn_helper, kernel_context, server, settings, websocket

os.environ["SERVER_SOFTWARE"] = "solara/" + str(solara.__version__)

logger = logging.getLogger("solara.server.flask")
blueprint = Blueprint("blueprint-solara", __name__)
websocket_extension = Sock()


class WebsocketWrapper(websocket.WebsocketWrapper):
    ws: simple_websocket.Server

    def __init__(self, ws: simple_websocket.Server) -> None:
        self.ws = ws
        self.lock = threading.Lock()

    def close(self):
        with self.lock:
            self.ws.close()

    def send_text(self, data: str) -> None:
        with self.lock:
            try:
                self.ws.send(data)
            except simple_websocket.ws.ConnectionClosed:
                raise websocket.WebSocketDisconnect()

    def send_bytes(self, data: bytes) -> None:
        with self.lock:
            try:
                self.ws.send(data)
            except simple_websocket.ws.ConnectionClosed:
                raise websocket.WebSocketDisconnect()

    async def receive(self):
        from anyio import to_thread

        try:
            return await to_thread.run_sync(lambda: self.ws.receive())  # type: ignore
        except simple_websocket.ws.ConnectionClosed:
            raise websocket.WebSocketDisconnect()


class ServerFlask(ServerBase):
    server: Any
    name = "flask"

    def __init__(self, port: int, host: str = "localhost", flask_app=None, url_prefix="", **kwargs):
        super().__init__(port, host, **kwargs)
        self.app = flask_app or app
        self.url_prefix = url_prefix
        self.server = None

    def has_started(self):
        return server.is_ready(f"http://{self.host}:{self.port}{self.url_prefix}")

    def signal_stop(self):
        if self.server:
            assert isinstance(self.server, HTTPServer)
            self.server.shutdown()  # type: ignore

    def serve(self):
        from werkzeug.serving import make_server

        self.server = make_server(self.host, self.port, self.app, threaded=True)  # type: ignore
        assert isinstance(self.server, HTTPServer)
        self.started.set()
        self.server.serve_forever(poll_interval=0.05)  # type: ignore


@blueprint.route("/jupyter/api/kernels/<id>")
def kernels(id):
    if not allowed():
        abort(401)
    return {"name": "lala", "id": "dsa"}


@websocket_extension.route("/jupyter/api/kernels/<kernel_id>/<name>")
def kernels_connection(ws: simple_websocket.Server, kernel_id: str, name: str):
    if not settings.main.base_url:
        settings.main.base_url = url_for("blueprint-solara.read_root", _external=True)
    if settings.oauth.private and not has_solara_enterprise:
        raise RuntimeError("SOLARA_OAUTH_PRIVATE requires solara-enterprise")
    if has_solara_enterprise:
        user = get_user()
        if user is None and settings.oauth.private:
            logger.error("app is private, requires login")
            ws.close(1008, "app is private, requires login")  # policy violation
            return
    else:
        user = None

    try:
        page_id = request.args["session_id"]
        session_id = request.cookies.get(server.COOKIE_KEY_SESSION_ID)
        logger.info("Solara kernel requested for session_id=%s kernel_id=%s", session_id, kernel_id)
        if session_id is None:
            logger.warning("no session cookie")
            session_id = "session-id-cookie-unavailable:" + str(uuid4())
        ws_wrapper = WebsocketWrapper(ws)
        headers_dict: Dict[str, List[str]] = {}
        for k, v in request.headers.__iter__():
            if k not in headers_dict.keys():
                headers_dict[k] = [v]
            else:
                headers_dict[k].append(v)
        asyncio.run(
            server.app_loop(
                ws_wrapper,
                cookies=request.cookies.to_dict(),
                headers=headers_dict,
                session_id=session_id,
                kernel_id=kernel_id,
                page_id=page_id,
                user=user,
            )
        )
    except:  # noqa
        logger.exception("Error in kernel handler")
        raise


@blueprint.route("/_solara/api/close/<kernel_id>", methods=["GET", "POST"])
def close(kernel_id: str):
    page_id = request.args["session_id"]
    context = kernel_context.contexts.get(kernel_id, None)
    if context is not None:
        context.page_close(page_id)
    return ""


@blueprint.route("/static/public/<path:path>")
def public(path):
    if not allowed():
        abort(401)
    directories = [app.directory.parent / "public" for app in appmod.apps.values()]
    for directory in directories:
        file = directory / path
        if file.exists():
            return send_from_directory(directory, path)
    return flask.Response("not found", status=404)


@blueprint.route("/static/assets/<path:path>")
def assets(path):
    if not allowed():
        abort(401)
    directories = server.asset_directories()
    for directory in directories:
        file = directory / path
        if file.exists():
            return send_from_directory(directory, path)
    return flask.Response("not found", status=404)


@blueprint.route("/jupyter/nbextensions/<dir>/<filename>")
def nbext(dir, filename):
    if not allowed():
        abort(401)
    for directory in server.nbextensions_directories:
        file = directory / dir / filename
        if file.exists():
            return send_from_directory(directory, dir + "/" + filename)
    return flask.Response("not found", status=404)


@blueprint.route("/static/<path:path>")
def serve_static(path):
    if not allowed():
        abort(401)
    return send_from_directory(server.solara_static, path)


if solara.settings.assets.proxy:

    @blueprint.route(f"/{cdn_helper.cdn_url_path}/<path:path>")
    def cdn(path):
        if not allowed():
            abort(401)
        cache_directory = settings.assets.proxy_cache_dir
        try:
            content = cdn_helper.get_data(Path(cache_directory), path)
        except PermissionError:
            return flask.Response("not found", status=404)
        mime = mimetypes.guess_type(path)
        return flask.Response(content, mimetype=mime[0])


@blueprint.route("/", defaults={"path": ""})
@blueprint.route("/<path:path>")
def read_root(path):
    root_path = url_for(".read_root")
    if root_path.endswith("/"):
        root_path = root_path[:-1]

    if not settings.main.base_url:
        settings.main.base_url = url_for("blueprint-solara.read_root", _external=True)

    session_id = request.cookies.get(server.COOKIE_KEY_SESSION_ID) or str(uuid4())
    if root_path:
        path = flask.request.path[len(root_path) :]
    content = server.read_root(path, root_path=root_path)
    if content is None:
        if not allowed():
            abort(401)
        return flask.Response("not found", status=404)

    if not allowed():
        return login()

    samesite = "lax"
    secure = False
    # we want samesite, so we can set a cookie when embedded in an iframe, such as on huggingface
    # however, samesite=none requires Secure https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Set-Cookie/SameSite
    # when hosted on the localhost domain we can always set the Secure flag
    # to allow samesite https://developer.mozilla.org/en-US/docs/Web/HTTP/Cookies#restrict_access_to_cookies
    o = urlparse(request.base_url)
    if request.headers.get("x-forwarded-proto", "http") == "https" or o.hostname == "localhost":
        samesite = "none"
        secure = True

    assert session_id is not None
    response = flask.Response(content, mimetype="text/html")
    response.set_cookie(server.COOKIE_KEY_SESSION_ID, value=session_id, secure=secure, samesite=samesite)
    return response


if has_solara_enterprise:
    blueprint.route("/_solara/auth/authorize")(authorize)
    blueprint.route("/_solara/auth/logout")(logout)
    blueprint.route("/_solara/auth/login")(login)


@blueprint.route("/readyz")
def readyz():
    json, status = server.readyz()
    return flask.Response(json, mimetype="application/json", status=status)


# using the blueprint and websocket blueprint makes it easier to integrate into other applications
websocket_extension.init_app(blueprint)
app = Flask(__name__, static_url_path="/_static")  # do not intervere with out static files
app.register_blueprint(blueprint)
if has_solara_enterprise:
    init_flask(app)

if __name__ == "__main__":
    app.run(debug=False, port=8765)

# we can only call this at the module level, which means that the solara script cannot import this
# module. This is a difference with the asgi standard, which provides a lifecycle hook (see starlette.py)
appmod.ensure_apps_initialized()
